import datetime
from functools import partial
import json.decoder
import math
import platform
import re

import matplotlib.pyplot as plt
import multiprocessing as mp
import os
import shutil
import subprocess
import sys
from argparse import ArgumentParser, RawTextHelpFormatter

from expam.run import die

EXPAM_VERSION = (0, 0, 7)
print("expam v%s" % ".".join(str(i) for i in EXPAM_VERSION))

# Check python version.
if sys.version_info[0] < 3:
    raise Exception("Must be using Python3.8 or later.")
elif sys.version_info[1] < 8:
    raise Exception("Must be using Python3.8 or later.")

# Determine current directory.
CWD = os.getcwd()

# Commands that don't require db_name parameter.
AUTO_COMMANDS = {
    "count",
    "from_jellyfish",
    "make_reads",
    "make_phylogeny",
}


#
#   Extract datetime point from string.
#
def get_time(data):
    date_data, time_data, am_pm = data.split(" ")[:3]

    mon, day, year = (int(v) for v in date_data.split("/"))
    hr, min, sec = (int(v) for v in time_data.split(":"))

    if "PM" in am_pm and hr < 12:
        hr += 12
    elif "AM" in am_pm and hr == 12:
        hr = 0

    return datetime.datetime(year, mon, day, hr, min, sec)


#
# Find the time of the first logging message from a set of log files at logs_dir.
#
def get_t0(logs_dir):
    from expam.sequences import ls

    def _first(itr):
        try:
            return itr[0]

        except IndexError:
            raise ValueError("Can't find main log file!")

    # Get effective t0 from main log.
    main_log_dir = _first([
        os.path.join(logs_dir, log_name)
        for log_name in ls(logs_dir, ext=".log")
        if "_main" in log_name
    ])

    with open(main_log_dir, "r") as f:
        log_data = f.readline()  # Get time of first log point.

    return main_log_dir, get_time(log_data)


#
# Request boolean confirmation from user.
#
def get_user_confirmation(msg):
    if "y/n" not in msg:
        msg += " (y/n)"

    while True:
        ans = input(msg)
        ans = ans.lower().strip()

        if ans == "y":
            return True
        elif ans == "n":
            return False
        else:
            die("Invalid response, please provide y/n.")


#
# Check that supplied parameters are integers, and if so
# return integer form.
#
def check_int_params(*params):
    for param in params:
        INVALID_PARAM_MSG = ("Invalid parameter (%s), must be integer!" % str(param))

        if param is not None:

            try:
                # Convert to float.
                param = float(param)

            except ValueError:
                die(INVALID_PARAM_MSG)

            # Convert to int and see if the value changes.
            new_param = int(param)
            if new_param != param:
                die(INVALID_PARAM_MSG)

            param = new_param

        yield param


#
# Try to return float form of arguments.
#
def check_float_params(*params):
    for param in params:
        INVALID_PARAM_MSG = ("Invalid parameter (%s), must be integer!" % str(param))

        if param is not None:
            try:
                param = float(param)

            except ValueError:
                die(INVALID_PARAM_MSG)

        yield param


def check_phylogeny(conf):
    if conf['phylogeny_path'] is None:
        return False
    return True


#
# Load a database configuration file.
#
def load_configuration_file(out_dir, return_config=False):
    from expam.run import JSONConfig

    # Load config file.
    try:
        configParser = JSONConfig(url=os.path.join(out_dir, "expam.conf"))
    except json.decoder.JSONDecodeError:
        die("Invalid configuration file!")
        return

    if return_config:
        return configParser

    genome_paths, phylogeny_path, k, n, pile_size = (
        configParser.get_paths(), configParser["phylogeny_path"],
        int(configParser["k"]), int(configParser["n"]), configParser["pile"])

    # Format phylogeny path.
    phylogeny_path = make_path_absolute(phylogeny_path, out_dir)

    return k, n, phylogeny_path, genome_paths, pile_size


#
# Calculate alpha to allow for at most one SNP. Assume a read length of 150.
#
def calculate_alpha(k):
    return 1e-3 * math.floor(((150 - k) / 150) * 1e3)


def _checker(cmd):
    _has_mash = shutil.which(cmd) is not None
    if not _has_mash:
        die('Could not find local installation of %s!' % cmd)


def _check_mash():
    _checker('mash')


def _check_mashtree():
    _checker('mashtree')


def _check_rapidnj():
    _checker('rapidnj')


def _check_quicktree():
    _checker('quicktree')


def subp(command, cwd=CWD):
    p = subprocess.run(
        command,
        shell=True,
        capture_output=True,
        cwd=cwd
    )

    # Check for fails.
    if p.returncode != 0:
        raise Exception("\n" + p.stderr.decode("utf-8"))

    return p.stdout.decode("utf-8")


def mash_sketch(k, s, p, sequences, out_dir):
    _check_mash()

    cmd_fmt = "mash sketch -k %d -p %d -s %d -o %s %s"
    cmd = cmd_fmt % (
        k,
        p,
        s,
        out_dir,
        ' '.join(sequences)
    )

    subp(cmd)


def sour_sketch(k, s, n, sequences, sig_dir):
    from expam.sourtree import make_signatures

    make_signatures(n, sequences, sig_dir, k, s)


def get_groups(conf, group_input):
    groups = conf.groups() if group_input is None else [group_input]

    # Check group has more than zero sequences.
    for group in groups:
        k, s, sequences = conf.group_get(group)

        if len(sequences) > 0:
            if k is None or s is None:
                raise ValueError("Parameters unspecified for group %s (k=%s, s=%s)!" % (group, str(k), str(s)))

            yield group


def check_sketches(conf, phy_dir, group, use_sourmash=False):
    sketch_dir = os.path.join(phy_dir, 'sketch')
    sketch_name_fmt = "%s.k%d.s%d.%s"

    if not os.path.exists(sketch_dir):
        return False

    for group_name in get_groups(conf, group):
        k, s, sequences = conf.group_get(group_name)
        file_name = sketch_name_fmt % (group_name, k, s, "%s")

        dest = os.path.join(sketch_dir, file_name % ("sour" if use_sourmash else "msh"))
        if not os.path.exists(dest):
            return False

    return True


def do_sketches(conf, phy_dir, group, use_sourmash=False):
    print("Sketching sequences...")

    sketch_dir = os.path.join(phy_dir, 'sketch')
    n_procs = conf['n'] if conf['n'] is not None else 1

    if not os.path.exists(sketch_dir):
        os.mkdir(sketch_dir)

    sketch_name_fmt = "%s.k%d.s%d.%s"

    for group_name in get_groups(conf, group):
        k, s, sequences = conf.group_get(group_name)
        file_name = sketch_name_fmt % (group_name, k, s, "%s")

        if use_sourmash:
            sig_path = os.path.join(sketch_dir, file_name % "sour")

            if not os.path.exists(sig_path):
                os.mkdir(sig_path)

            sour_sketch(k=k, s=s, n=n_procs, sequences=sequences, sig_dir=sig_path)

        else:
            _check_mash()

            # Use mash by default.
            file_path = os.path.join(sketch_dir, file_name % "msh")
            mash_sketch(k=k, s=s, p=n_procs, sequences=sequences, out_dir=file_path)


def check_distances(conf, phy_dir, group):
    sketch_dir = os.path.join(phy_dir, 'sketch')
    dist_dir = os.path.join(phy_dir, 'distance')

    if not os.path.exists(dist_dir):
        return False

    dist_name_fmt = "%s.k%d.s%d.tab"

    for group_name in get_groups(conf, group):
        k, s, sequences = conf.group_get(group_name)
        file_name = dist_name_fmt % (group_name, k, s)

        dest = os.path.join(sketch_dir, file_name)
        if not os.path.exists(dest):
            return False

    return True


def do_distances(conf, phy_dir, group, use_sourmash=False):
    print("Calculating pairwise distances...")

    sketch_dir = os.path.join(phy_dir, 'sketch')
    dist_dir = os.path.join(phy_dir, 'distance')
    n_procs = conf['n'] if conf['n'] is not None else 1

    if not os.path.exists(dist_dir):
        os.mkdir(dist_dir)

    dist_name_fmt = "%s.k%d.s%d.%s"

    for group_name in get_groups(conf, group):
        k, s, sequence_files = conf.group_get(group_name)

        sketch_name = dist_name_fmt % (group_name, k, s, '%s')

        matrix_name = dist_name_fmt % (group_name, k, s, 'tab')
        matrix_file = os.path.join(dist_dir, matrix_name)

        if use_sourmash:
            sketch_dir = os.path.join(sketch_dir, sketch_name % 'sour')
            sour_dist(n=n_procs, sig_dir=sketch_dir, matrix_dir=matrix_file)

        else:
            _check_mash()

            sketch_file = os.path.join(sketch_dir, sketch_name % 'msh')
            mash_dist(n_procs, sketch_file, matrix_file)


def check_trees(conf, phy_dir, group):
    tree_dir = os.path.join(phy_dir, 'tree')

    if not os.path.exists(tree_dir):
        return False

    for group_name in get_groups(conf, group):
        tree_name = '%s.nwk' % group_name
        tree_path = os.path.join(tree_dir, tree_name)

        if not os.path.exists(tree_path):
            return False

    return True


def do_trees(conf, phy_dir, group, use_quicktree=False):
    print("Running neighbour-joining...")

    dist_dir = os.path.join(phy_dir, 'distance')
    tree_dir = os.path.join(phy_dir, 'tree')

    n_procs = conf['n'] if conf['n'] is not None else 1
    dist_fmt = "%s.k%d.s%d.tab"

    if not os.path.exists(tree_dir):
        os.mkdir(tree_dir)

    for group_name in get_groups(conf, group):
        k, s, _ = conf.group_get(group_name)

        matrix_name = dist_fmt % (group_name, k, s)
        matrix_path = os.path.join(dist_dir, matrix_name)

        tree_name = "%s.nwk" % group_name
        tree_path = os.path.join(tree_dir, tree_name)

        if use_quicktree:
            quicktree(matrix_dir=matrix_path, tree_dir=tree_path)

        else:
            rapidnj(matrix_dir=matrix_path, n=n_procs, tree_dir=tree_path)


def mash_dist(p, sketch_dir, matrix_dir):
    _check_mash()

    cmd_fmt = "mash dist -p %d -t %s %s"
    cmd = cmd_fmt % (
        p,
        sketch_dir,
        sketch_dir
    )

    _unfmt_dst = subp(cmd)
    lines = [line.strip() for line in _unfmt_dst.split('\n')]

    names = lines[0].split('\t')
    lines[0] = '%d' % len(names[1:])

    # Format sequences names from within directories.
    for i in range(1, len(lines)):
        line = lines[i].split('\t')
        line[0] = os.path.basename(line[0])
        lines[i] = '\t'.join(line)

    with open(matrix_dir, 'w') as f:
        f.write('\n'.join(lines))


def sour_dist(n, sig_dir, matrix_dir):
    from expam.sourtree import make_distances

    make_distances(sig_dir, n, matrix_dir)


def _format_tree_string(nwk):
    seq_types = [".fna", ".faa"]
    comp_types = [".tar.gz", ".gz"]

    nwk = nwk.replace("'", "")

    for seq_type in seq_types:
        for comp_type in comp_types:
            ext = seq_type + comp_type + ":"
            nwk = nwk.replace(ext, ":")

    return nwk


def rapidnj(matrix_dir, n, tree_dir):
    _check_rapidnj()

    cmd_fmt = "rapidnj %s -i pd -o t -c %d"
    cmd = cmd_fmt % (matrix_dir, n)

    unformatted_tree = subp(cmd)
    tree = _format_tree_string(unformatted_tree)

    with open(tree_dir, 'w') as f:
        f.write(tree)

    print("RapidNJ wrote tree to %s." % tree_dir)


def quicktree(matrix_dir, tree_dir):
    _check_quicktree()

    cmd_fmt = "quicktree -in m -out t %s"
    cmd = cmd_fmt % matrix_dir

    unformatted_tree = subp(cmd)
    tree = _format_tree_string(unformatted_tree)

    with open(tree_dir, 'w') as f:
        f.write(tree)

    print("QuickTree wrote tree to %s." % tree_dir)


def do_mashtree(conf, phy_dir, group):
    print("Creating mashtree...")
    _check_mashtree()

    tree_dir = os.path.join(phy_dir, 'tree')
    temp_dir = os.path.join(phy_dir, 'tmp')

    n_procs = conf['n'] if conf['n'] is not None else 1

    if not os.path.exists(tree_dir):
        os.mkdir(tree_dir)

    for group_name in get_groups(conf, group):
        k, s, sequences = conf.group_get(group_name)
        tree_path = os.path.join(tree_dir, "%s.nwk" % group_name)

        mashtree(k, n_procs, s, sequences, tree_path, temp_dir)


def mashtree(k, n, s, sequences, tree_dir, temp_dir):
    _names_file = os.path.join(temp_dir, 'sequence_names.txt')

    def delete_temp():
        nonlocal temp_dir

        shutil.rmtree(temp_dir)
        print("Deleted temporary directory %s." % temp_dir)

    # Make temporary directory.
    try:
        print("Making temporary directory %s..." % temp_dir)
        os.mkdir(temp_dir)

    except OSError:
        try:
            delete_temp()
        except FileNotFoundError:
            pass

        die("Could not create temporary directory!")

    # Write file of files.
    try:
        with open(_names_file, "w") as f:
            f.write("\n".join(sequences))

        # Make mashtree and insert into configuration file.
        print("Making mashtree...")

    except OSError:
        die("Could not write genome paths to text file!")

    cmd = "mashtree --numcpus %d --kmerlength %d --sketch-size %d --file-of-files %s" \
          % (n, k, s, _names_file)

    # Build the tree.
    try:
        return_val = subp(cmd, cwd=temp_dir)

        with open(tree_dir, "w") as f:
            f.write(return_val)

        print("Phylogeny created in %s." % tree_dir)

    except:
        try:
            delete_temp()
        except FileNotFoundError:
            pass

        raise

    # Delete the temporary directory.
    try:
        delete_temp()

    except OSError:
        die("Could not delete temporary directory %s!" % temp_dir)


def finalise_tree(conf, phy_dir, db_name):
    print("Finalising tree...")

    from expam.sequences import ls

    tree_dir = os.path.join(phy_dir, 'tree')
    tree_name = "%s.nwk" % db_name
    tree_path = os.path.join(tree_dir, tree_name)

    tree_files = [os.path.join(tree_dir, name) for name in ls(tree_dir, ext='.nwk') if name != tree_name]

    def get_tree_data(file_url):
        with open(file_url, 'r') as f:
            tree_data = f.read().strip().rstrip(";")

        return tree_data

    if len(tree_files) == 1:
        nwk_data = get_tree_data(tree_files[0]) + ";"

        with open(tree_path, 'w') as f:
            f.write(nwk_data)

    else:
        try:
            template = get_tree_data(tree_path) + ";"

        except FileNotFoundError:
            raise Exception("No template found! Please write a template to %s!" % tree_path)

        template_groups = re.findall(r"{{(\S+?)}}", template)

        tree_data = {
            group_name: get_tree_data(tree_file)
            for group_name in conf.groups()
            for tree_file in tree_files
            if group_name in tree_file
        }

        for template_group in template_groups:
            if template_group not in tree_data:
                die("Template group %s has no tree!" % template_group)

            else:
                template = template.replace("{{%s}}" % template_group, tree_data[template_group])

        with open(tree_path, 'w') as f:
            f.write(template)

    return tree_path


#
# Ensure that a given path is absolute.
#
def make_path_absolute(dir, base_dir):
    if base_dir not in dir:
        dir = os.path.join(base_dir, dir)

    return dir


#
# Initialise a database in the current folder.
#
def initialise_db(db_name):
    from expam.run import JSONConfig

    path = os.path.join(CWD, db_name)
    conf_path = os.path.join(path, "expam.conf")

    database_path = os.path.join(path, "database")
    phylogeny_path = os.path.join(path, "phylogeny")
    logs_path = os.path.join(path, "logs")
    results_path = os.path.join(path, "results")

    # Create the required folders.
    try:
        os.makedirs(phylogeny_path)

    except FileExistsError:
        # Check if the main folder already exists.
        if os.path.exists(conf_path):
            try:
                _conf = JSONConfig(url=conf_path)
                die("Database already exists!")

            except json.decoder.JSONDecodeError:
                die("Invalid name: non-database folder exists!")

        else:
            # Database has been initialised, but missing configuration file.
            # Get user confirmation this is indeed a database directory.
            is_db = get_user_confirmation("Database folder appears damaged. Restart it? (y/n)")
            if not is_db:
                die("Aborting...")

    except OSError:
        raise Exception("Failed to make directories %s!" % phylogeny_path)

    else:
        print("Successfully created directories %s!" % phylogeny_path)

    # Initialise the configuration file.
    conf = JSONConfig(name=db_name)
    conf.save(conf_path)
    print("Fresh database configuration generated!")

    # Create folder for logs.
    while True:
        try:
            os.mkdir(logs_path)
            print("Logs path %s created!" % logs_path)

            break

        except FileExistsError:
            print("Logs path %s already exists. Cleaning..." % logs_path)

            shutil.rmtree(logs_path)
            print("Previous logs cleared!")

    # Create database and results path.
    for db_path in (database_path, results_path):
        try:
            os.mkdir(db_path)

            print("Made path %s." % db_path)

        except FileExistsError:
            continue


#
# Modify sequences in database configuration file.
#
def _modify(directory, conf_path, first_n=None, group=None, add=True):
    from expam.run import JSONConfig
    from expam.sequences import ls

    # Ensure absolute path.
    base_url = make_path_absolute(
        dir=directory,
        base_dir=CWD
    )

    # Get list of sequences in base_url.
    seq_names = ls(base_url)

    if (first_n is None) or (first_n > len(seq_names)):
        first_n = len(seq_names)

    seq_dirs = [
        os.path.join(base_url, seq_names[i])
        if seq_names[i] != base_url else seq_names[i]
        for i in range(first_n)
    ]

    # Append directory to database configuration.
    conf = JSONConfig(url=conf_path)

    if add:
        conf.add_sequence(seq_dirs, group=group)
        print("Added %d files from %s." % (len(seq_dirs), base_url))
    else:
        conf.remove_sequence(seq_dirs, group=group)
        print("Removed %d files from %s." % (len(seq_dirs), base_url))

    conf.save()


def add_sequences(directory, conf_path, first_n=None, group=None):
    _modify(directory, conf_path, first_n, group, add=True)


def remove_sequences(directory, conf_path, first_n=None, group=None):
    _modify(directory, conf_path, first_n, group, add=False)


#
# Start building database with current configuration.
#
def build_database(db_path, plot):
    from expam.run import main as expam_main, JSONConfig, Timer

    conf_path = os.path.join(db_path, "expam.conf")
    conf = JSONConfig(conf_path)

    # Check if a phylogeny has been provided.
    if not check_phylogeny(conf):
        do_mashtree(conf, phy_dir=os.path.join(db_path, 'phylogeny'), group=None)

    # Read configuration file.
    k, n, phylogeny, genome_paths, pile = load_configuration_file(db_path)

    # Ensure correct phylogeny path.
    phylogeny = make_path_absolute(
        dir=phylogeny,
        base_dir=db_path
    )

    clear_logs(os.path.join(db_path, "logs"))

    # Set building database.
    try:
        with Timer() as t:
            expam_main(
                out_dir=db_path,
                genome_paths=genome_paths,
                phylogeny_path=phylogeny,
                k=k,
                n=n - 1,  # Account for main process.
                pile_size=pile,
                plot=plot
            )

        print("expam: " + str(t))

    finally:
        print(f'\nPID - {os.getpid()} dying...')


#
# Delete old log files.
#
def clear_logs(log_path):
    print("Clearing old log files...")

    try:
        shutil.rmtree(log_path)

    except FileNotFoundError:
        pass

    try:
        os.mkdir(log_path)

    except OSError:
        die("Can't make log path %s!" % log_path)


#
# Importable main.
#
def main():
    # Set process creation method.
    if platform.system() != "Windows":
        mp.set_start_method("fork")

    # Set command line arguments.
    parser = ArgumentParser(description="  expam CLI\n--------------\n", formatter_class=RawTextHelpFormatter)
    parser.add_argument("command", default=None,
                        help='\nCommand to execute. Valid commands include:\n'
                             '-------------------------------------------\n'
                             'create:-\tInitialise database.\n'
                             'build:-\t\tStart building database.\n'
                             'print:-\t\tPrint current database parameters.\n'
                             'run:-\t\tRun reads against database.\n'
                             'add:-\t\tAdd sequence to the database.\n'
                             'remove:-\tRemove sequence from database (only impacts future db builds).\n'
                             'set:-\t\tSet database build parameters.\n'
                             'mashtree:-\tCreate mashtree from current sequences and add to database.\n'
                             'quickrun:-\tInitialise, set parameters and start building db (assumes\n'
                             '\t\t\tsequences all lie in the same folder).\n'
                             'make_reads:-\tUniformly sample reads of length l from some input sequence.\n'
                             '\t\tThis is for testing purposes only, and is not a replacement\n'
                             '\t\tfor actual read generating software.\n',
                        metavar="[command]")
    parser.add_argument("-db", "--db_name", dest="db_name",
                        help="Name of database.",
                        metavar="[database name]")
    parser.add_argument("-k", "--kmer", dest="k",
                        help="Length of mer used for analysis.",
                        metavar="[k value (int)]")
    parser.add_argument("-n", "--n_processes", dest="n",
                        help="Number of CPUs to use for processing.",
                        metavar="[n (int)]")
    parser.add_argument("-s", "--sketch", dest="sketch",
                        help="Sketch size for mash.",
                        metavar="[sketch size (int)]")
    parser.add_argument("-p", "--phylogeny", dest="phylogeny",
                        help="URL of Newick file containing phylogeny.",
                        metavar="[phylogeny URL]")
    parser.add_argument("-d", "--directory", dest="directory",
                        help="File URL, context depending on command supplied.",
                        metavar="[directory]")
    parser.add_argument("-l", "--length", dest="length",
                        help="Length of simulated reads.",
                        metavar="[read length]")
    parser.add_argument("-o", "--out", dest="out_url",
                        help="URL to save sequences.",
                        metavar="[out URL]")
    parser.add_argument("-y", "--pile", dest="pile",
                        help="Number of genomes to pile at a time (or inf).",
                        metavar="[pile size]")
    parser.add_argument("-e", "--error_rate", dest="error_rate",
                        help="Generate error in reads (error ~ reads with errors / reads).",
                        metavar="[error rate]")
    parser.add_argument("-t", "--truth", dest="truth_dir",
                        help="Location of truth dataset.")
    parser.add_argument("--plot", dest="plot", default=False, action="store_true",
                        help="Plot timing data of database build.")
    parser.add_argument("--first", dest="first_n", default=None,
                        help="Add first n genomes in folder.")
    parser.add_argument("--cutoff", dest="cutoff", default=0,
                        help="Ignore organisms with less than `cutoff` reads in results.")
    parser.add_argument("--cpm", dest="cpm", default=100,
                        help="Counts/million cutoff for read-count to be non-negligible.")
    parser.add_argument("--taxonomy", dest="taxonomy", default=False, action="store_true",
                        help="Convert phylogenetic results to taxonomic results.")
    parser.add_argument("--phyla", dest="phyla", default=False, action="store_true",
                        help="Colour phylotree results by phyla.")
    parser.add_argument("--rank", dest="rank", default=None,
                        help="Rank at which to sort results.")
    parser.add_argument("--keep_zeros", dest="keep_zeros", default=False, action="store_true",
                        help="Keep nodes of output where no reads have been assigned.")
    parser.add_argument("--ignore_names", dest="ignore_node_names", default=False, action="store_true")
    parser.add_argument("--group", dest="groups", action="append", nargs="+",
                        help="Space-separated list of sample files to be treated as a single group in phylotree.")
    parser.add_argument("--colour_list", dest="colour_list", nargs="+",
                        help="List of colours to use when plotting groups in phylotree.")
    parser.add_argument("--name", dest="results_name", default=None,
                        help="Name of results folder.")
    parser.add_argument("--circle_scale", dest="circle_scale", default=1.0,
                        help="Scale of circles that represent splits in phylotree.")
    parser.add_argument("--sourmash", dest="use_sourmash", default=False, action="store_true",
                        help="Use sourmash for distance estimation.")
    parser.add_argument("--rapidnj", dest="use_rapidnj", default=True, action="store_true",
                        help="Use RapidNJ for Neighbour-Joining algorithm.")
    parser.add_argument("--quicktree", dest="use_quicktree", default=False, action="store_true",
                        help="Use QuickTree for Neighbour-Joining algorithm.")
    parser.add_argument("--paired", dest="paired_end", default=False, action="store_true",
                        help="Treat reads as paired-end.")
    parser.add_argument("--alpha", dest="alpha", default=None,
                        help="Percentage requirement for classification subtrees (see Tutorials 1 & 2).")

    # Parse arguments.
    args = parser.parse_args()

    # Argument groups.
    runtime_args = args.command, args.db_name, args.k, args.n, args.phylogeny, args.alpha
    directory_args = args.directory, args.out_url, args.truth_dir
    param_args = args.length, args.pile, args.error_rate, args.first_n, args.sketch, args.paired_end
    summary_args = args.plot, args.cutoff, args.cpm, args.taxonomy, args.results_name
    plot_args = args.groups, args.phyla, args.keep_zeros, not args.ignore_node_names, args.colour_list, \
                args.circle_scale, args.rank
    tree_args = args.use_sourmash, args.use_rapidnj, args.use_quicktree

    command, db_name, k, n, phylogeny, alpha = runtime_args
    directory, out_url, truth_dir = directory_args
    length, pile_size, error_rate, first_n, sketch, paired_end = param_args
    plot, cutoff, cpm, taxonomy, results_name = summary_args
    groups, plot_phyla, keep_zeros, use_node_names, colour_list, circle_scale, at_rank = plot_args
    use_sourmash, use_rapidnj, use_quicktree = tree_args

    group = None if groups is None else groups[0][0]  # When referring to sequence groups, not plotting groups.

    # Check we have a database name.
    if db_name is None and command not in AUTO_COMMANDS:
        try:
            DB_DIR = os.environ["EXPAM_DB_DIR"]
            db_name = os.path.basename(DB_DIR)

        except KeyError:
            die("A database name is required!")
            return
    else:
        DB_DIR = make_path_absolute(str(db_name), CWD)

    # Check we have a command.
    if command is None:
        die("No command supplied!")

    # Check valid form of passed parameters.
    k, n, length, first_n, cutoff = check_int_params(k, n, length, first_n, cutoff)
    error_rate, cpm, circle_scale = check_float_params(error_rate, cpm, circle_scale)

    # Determine supposed configuration file directory.
    CONF_DIR = os.path.join(DB_DIR, "expam.conf")
    PHY_DIR = os.path.join(DB_DIR, "phylogeny")

    def is_hex(string):
        if string[0] != "#" or len(string) != 7:
            return False

        try:
            int(string[1:].upper(), 16)
        except TypeError:
            return False

        return True

    # Format groups.
    if groups is not None:
        groups = [v for v in groups if v]

        for i in range(len(groups)):
            if is_hex(groups[i][0]):
                groups[i] = (groups[i][0], tuple(groups[i][1:]))
            else:
                groups[i] = (None, tuple(groups[i]))

    # Check colour list.
    if colour_list is not None:
        colour_list = [hex for hex in colour_list if is_hex(hex)]

        if not colour_list:
            colour_list = None

    # If plotting phyla, check we have taxonomic information.
    if (not plot_phyla) and (at_rank is None):
        name_to_lineage = None

    else:
        try:
            from expam.classification import load_taxonomy_map

            name_to_lineage, name_to_rank = load_taxonomy_map(DB_DIR)
        except FileNotFoundError:
            die("Run command `download_taxonomy` first to collect taxa for your genomes!")

            return

    # Command loop.

    #
    # Create and set configuration, then build database all in one step.
    #   *** Assumes all reference sequences are located in the same folder.
    #
    if command == "quickrun":
        from expam.run import JSONConfig

        # Initialise database folders.
        initialise_db(db_name)

        # Set the requested database parameters.
        conf = JSONConfig(url=CONF_DIR)
        conf.set(k=k, n=n, phylogeny_path=phylogeny, pile=pile_size)
        conf.save()

        # Add all sequences in the provided folder.
        add_sequences(directory, CONF_DIR, first_n)

        # Start building database.
        build_database(DB_DIR, plot)

    #
    # Print setup for default environment variable.
    #
    elif command == "default_db":
        print("export EXPAM_DB_DIR=%s" % DB_DIR)


    #
    # Initialise a database.
    #
    elif command == "create":
        initialise_db(db_name)

    #
    # Build database with set parameters.
    #
    elif command == "build":
        build_database(DB_DIR, plot)

    #
    # Print current parameter values in configuration file.
    #
    elif command == "print":
        # Check database exists.
        if not os.path.exists(CONF_DIR):
            die("'%s' not a database!" % db_name)

        config = load_configuration_file(DB_DIR, return_config=True)
        print(config)

    #
    # Run database against some reads.
    #
    elif command == "run":
        from expam.classification import run_classifier

        logs_path = os.path.join(DB_DIR, 'logs')
        clear_logs(logs_path)

        # Read configuration file.
        config = load_configuration_file(DB_DIR, return_config=True)

        k, n, phylogeny_path = int(config["k"]), int(config["n"]), config["phylogeny_path"]
        keys_shape, values_shape = tuple(config["keys_shape"]), tuple(config["values_shape"])

        # Correct for trimmed keys_shape.
        if len(keys_shape) == 1:
            keys_shape = keys_shape + (1,)

        phylogeny_path = make_path_absolute(phylogeny_path, DB_DIR)

        # Check alpha parameter is defined.
        alpha = calculate_alpha(k) if alpha is None else float(alpha)

        # Run expam classification.
        run_classifier(
            reads_url=directory,
            out_dir=DB_DIR,
            k=k,
            n=n - 1,  # Account for main process.
            phylogeny_path=phylogeny_path,
            keys_shape=keys_shape,
            values_shape=values_shape,
            logging_dir=logs_path,
            taxonomy=taxonomy,
            cutoff=cutoff,
            groups=groups,
            cpm=cpm,
            use_node_names=use_node_names,
            phyla=plot_phyla,
            name_taxa=name_to_lineage,
            colour_list=colour_list,
            results_name=results_name,
            circle_scale=circle_scale,
            paired_end=paired_end,
            alpha=alpha
        )

    #
    # Download mapping from accession ids to taxonomy.
    #
    elif command == "download_taxonomy":
        from expam.classification import accession_to_taxonomy

        # Convert accession ids to taxonomic lineages.
        accession_to_taxonomy(DB_DIR)

    #
    # Convert phylogenetic results to taxonomic results.
    #
    elif command == "to_taxonomy":
        if results_name is None:
            die("Require results directory (--name)!")

        from expam.classification import TAXID_LINEAGE_MAP_NAME, PHY_RESULTS, TAX_RESULTS, \
            load_taxonomy_map, name_to_id, ClassificationResults

        all_results_dir = os.path.join(DB_DIR, "results")
        results_dir = os.path.join(all_results_dir, results_name) \
            if results_name[:len(all_results_dir)] != all_results_dir else results_name

        if not os.path.exists(results_dir):
            die("Could not find results %s!" % results_dir)

        map_url = os.path.join(DB_DIR, TAXID_LINEAGE_MAP_NAME)
        if not os.path.exists(map_url):
            die("Run command `download_taxonomy` first to collect taxa for your genomes!")

        config = load_configuration_file(DB_DIR, return_config=True)
        phylogeny_path = config["phylogeny_path"]
        phylogeny_path = make_path_absolute(phylogeny_path, DB_DIR)

        index, phylogenyIndex = name_to_id(phylogeny_path)
        name_to_lineage, taxon_to_rank = load_taxonomy_map(DB_DIR)

        # Load phylogenetic results.
        phy_results_url = os.path.join(results_dir, PHY_RESULTS)
        tax_results_path = os.path.join(results_dir, TAX_RESULTS)

        if not os.path.exists(tax_results_path):
            os.mkdir(tax_results_path)

        results = ClassificationResults(
            index=index,
            phylogeny_index=phylogenyIndex,
            in_dir=phy_results_url,
            out_dir=results_dir,
            groups=groups,
            keep_zeros=keep_zeros,
            cutoff=cutoff,
            cpm=cpm,
            use_node_names=use_node_names,
            phyla=plot_phyla,
            name_taxa=name_to_lineage,
            colour_list=colour_list,
            circle_scale=circle_scale
        )
        results.to_taxonomy(name_to_lineage, taxon_to_rank, tax_results_path)
        results.draw_results()

    #
    # Plot tree.
    #
    elif command == "draw_tree":
        from ete3 import AttrFace, faces, Tree, TreeStyle, NodeStyle
        from expam.classification import DISC_COLOURS, PHYLA_COLOURS

        config = load_configuration_file(DB_DIR, return_config=True)
        phylogeny_path = config["phylogeny_path"]

        with open(make_path_absolute(phylogeny_path, DB_DIR), "r") as f:
            newick_string = f.read().strip()

        tree = Tree(newick_string, format=1)
        rank_colours = {}

        def layout(node):
            nonlocal name_to_lineage, name_to_rank

            if node.is_leaf():
                if use_node_names:
                    faces.add_face_to_node(AttrFace("name", fsize=20), node, column=0, position="aligned")

                if at_rank is not None:
                    lineage = name_to_lineage[node.name]

                    for name in lineage:
                        rank = name_to_rank[name][1]

                        if rank == at_rank:
                            if name not in rank_colours:
                                try:
                                    rank_colours[name] = DISC_COLOURS.pop()
                                except IndexError:
                                    die("Too many members of that rank!")

                            node_style = NodeStyle()
                            node_style['bgcolor'] = rank_colours[name]
                            node.set_style(node_style)

                elif plot_phyla:
                    for phyla, phyla_colour in PHYLA_COLOURS:
                        lineage = name_to_lineage[node.name]

                        if phyla in lineage:
                            node_style = NodeStyle()
                            node_style['bgcolor'] = phyla_colour
                            node.set_style(node_style)

        # Print tree to file.
        ts = TreeStyle()
        ts.mode = "c"
        ts.show_leaf_name = False
        ts.layout_fn = layout
        ts.force_topology = True
        ts.allow_face_overlap = False
        ts.draw_guiding_lines = True
        ts.root_opening_factor = 1

        phy_path = "phylotree.pdf" if not out_url else os.path.join(out_url)
        tree.render(
            phy_path,
            tree_style=ts,
            w=4000,
            units="px"
        )
        print("Phylogenetic tree written to %s!" % phy_path)

    #
    # Calculate precision and recall.
    #
    elif command == "evaluate":
        if not os.path.exists(directory):
            die("Couldn't find results path %s!" % directory)

        if not os.path.exists(truth_dir):
            die("Couldn't find truth dataset %s!" % truth_dir)

        import numpy as np
        import pandas as pd

        from expam.classification import load_taxonomy_map, request_labels, RANK_ORDER

        # Performance statistics.
        x = ['phylum', 'class', 'order', 'family', 'genus', 'species']
        y_prec_lib = []
        y_prec_cons = []
        y_rec_lib = []
        y_rec_cons = []

        taxid_to_lineage, taxon_to_rank = load_taxonomy_map(DB_DIR, convert_to_name=False)

        # Classifier results.
        expam_df = pd.read_csv(directory, sep="\t", names=['taxid', 'c_perc', 'c_cumul', 'c_count',
                                                           's_perc', 's_cumul', 's_count', 'rank', 'lineage'])
        expam_df['t_cumul'] = expam_df['s_cumul'] + expam_df['c_cumul']

        # Ensure we have representation for each rank in x.
        lineages = list({
            lineage
            for taxid, lineage in taxid_to_lineage.items()
            if taxid in expam_df['taxid']
        })

        for lineage in lineages:
            current_ranks = [(i, RANK_ORDER.index(taxon_to_rank[child][1])) for i, child in enumerate(lineage)]
            current_rank_indices = [rank_index for ind, rank_index in current_ranks]
            missing_rank_indices = [RANK_ORDER.index(rank) for rank in x if rank not in current_rank_indices]

            for rank_index in missing_rank_indices:
                # Find the closest child.
                child_index = min([ind for ind, child_rank_index in current_ranks
                                   if child_rank_index > rank_index])

                # Add this closest child as the representative member for this missing rank.
                child_taxid = taxon_to_rank[lineage[child_index]][0]
                child_df_index = expam_df.index[expam_df['taxid'] == child_taxid]

                new_row = expam_df.loc[child_df_index].copy()
                new_row['rank'] = RANK_ORDER[rank_index]

                # Append row to dataframe.
                expam_df.append(new_row, ignore_index=True)

        # Truth set.
        truth_df = pd.read_csv(truth_dir, sep="\t", names=['taxid', 'counts', 'abundance', 'rank', 'scientific name'])
        truth_df['taxid'] = truth_df['taxid'].apply(str)

        # Collect lineages for each taxid.
        species_taxids = truth_df['taxid'].tolist()
        taxid_to_lineage, taxon_to_rank = request_labels("taxonomy", "xml", species_taxids)

        taxid_to_lineage = {
            taxid: lineage.split(",")
            for taxid, lineage in taxid_to_lineage
        }
        taxon_to_rank = {
            taxon: rank.split(",")
            for taxon, rank in taxon_to_rank.items()
        }

        def at_rank(lineage, rank):
            nonlocal taxon_to_rank, RANK_ORDER

            z = RANK_ORDER.index(rank)
            descendants = [
                taxon_to_rank[name]
                for name in lineage
                if RANK_ORDER.index(taxon_to_rank[name][1]) >= z
            ]

            return descendants[0][0]

        rank_taxids = {
            rank: list({at_rank(lineage, rank) for lineage in taxid_to_lineage.values()})
            for rank in x
        }

        summary = np.sum if not taxonomy else len

        for rank, taxa in rank_taxids.items():
            df_subset = expam_df[expam_df['taxid'].isin(taxa)]

            # Total number of reads classified at given rank.
            mask = (expam_df['rank'] == rank) & (expam_df['c_cumul'] > cutoff)
            mask_comb = (expam_df['rank'] == rank) & (expam_df['t_cumul'] > cutoff)

            n_classified = summary(expam_df[mask]['c_cumul'])
            n_classified_combined = summary(expam_df[mask_comb]['t_cumul'])

            n_total = n if not taxonomy else len(taxa)

            # Precision and recall for confidently classified.
            y_prec_cons.append(
                summary(
                    df_subset[df_subset['c_cumul'] > cutoff]['c_cumul']
                ) / n_classified)
            y_rec_cons.append(
                summary(
                    df_subset[df_subset['c_cumul'] > cutoff]['c_cumul']
                ) / n_total)

            # Precision and recall under liberal classification regime.
            y_prec_lib.append(
                summary(
                    df_subset[df_subset['t_cumul'] > cutoff]['t_cumul']
                ) / n_classified_combined)
            y_rec_lib.append(
                summary(
                    df_subset[df_subset['t_cumul'] > cutoff]['t_cumul']
                ) / n_total)

        fig, (prec_ax, rec_ax) = plt.subplots(1, 2, figsize=(20, 10))

        prec_ax.set_title("Precision")
        prec_ax.plot(x, y_prec_lib, label="liberal", color="green", marker='o')
        prec_ax.plot(x, y_prec_cons, label="conservative", color="blue", marker='o')
        prec_ax.set_ylim([-0.05, 1.05])
        prec_ax.grid()

        rec_ax.set_title("Recall")
        rec_ax.plot(x, y_rec_lib, label="liberal", color="green", marker='o')
        rec_ax.plot(x, y_rec_cons, label="conservative", color="blue", marker='o')
        rec_ax.set_ylim([-0.05, 1.05])
        rec_ax.grid()

        prec_ax.legend(loc="best")
        rec_ax.legend(loc="best")

        # Save figure to disk.
        fig.savefig(os.path.join(out_url, "prec_recall.png"))

        # Save raw counts to disk.
        df = pd.DataFrame(index=x)

        df['prec_lib'] = y_prec_lib
        df['prec_cons'] = y_prec_cons
        df['rec_lib'] = y_rec_lib
        df['rec_cons'] = y_rec_cons
        df.to_csv(os.path.join(out_url, "prec_recall.csv"), sep="\t")

    #
    # Add sequence to configuration file.
    #
    elif command == "add":
        add_sequences(directory, CONF_DIR, group=group, first_n=first_n)

    #
    # Remove sequence from configuration file.
    #
    elif command == "remove":
        remove_sequences(directory, CONF_DIR, group=group, first_n=first_n)

    #
    # Set parameters in configuration file.
    #
    elif command == "set":
        from expam.run import JSONConfig

        conf = JSONConfig(url=CONF_DIR)

        if group is None:
            conf.set(k=k, n=n, phylogeny_path=phylogeny, pile=pile_size, s=sketch)
        else:
            conf.group_set(group_name=group, k=k, s=sketch)

        conf.save()

    #
    # Sketch sequences.
    #
    elif command == "sketch":
        from expam.run import JSONConfig

        conf = JSONConfig(url=CONF_DIR)
        do_sketches(conf, PHY_DIR, group, use_sourmash=use_sourmash)


    #
    # Compute distance tree.
    #
    elif command == "distance":
        from expam.run import JSONConfig

        conf = JSONConfig(url=CONF_DIR)
        do_distances(conf, PHY_DIR, group, use_sourmash=use_sourmash)

    #
    # Execute neighbour joining on tree.
    #
    elif command == "nj":
        from expam.run import JSONConfig

        conf = JSONConfig(url=CONF_DIR)
        do_trees(conf, PHY_DIR, group, use_quicktree=use_quicktree)

    #
    # Build phylogeny.
    #
    elif command == "tree":
        from expam.run import JSONConfig

        def argmax(*checks):
            if not any(checks):
                return 0

            return max(i for i, val in enumerate(checks + (True, )) if val is True)

        conf = JSONConfig(url=CONF_DIR)

        entry_points = (
            partial(do_sketches, conf=conf, phy_dir=PHY_DIR,
                    group=group, use_sourmash=use_sourmash),    # Sketch sequences.
            partial(do_distances, conf=conf, phy_dir=PHY_DIR,
                    group=group, use_sourmash=use_sourmash),    # Pairwise distances.
            partial(do_trees, conf=conf, phy_dir=PHY_DIR,
                    group=group, use_quicktree=use_quicktree)   # NJ trees.
        )
        entry_stage = argmax(
            check_sketches(conf, PHY_DIR, group, use_sourmash=use_sourmash),
            check_distances(conf, PHY_DIR, group),
            check_trees(conf, PHY_DIR, group)
        )

        for stage in range(entry_stage, 3):
            entry_points[stage]()

        tree_dir = finalise_tree(conf, PHY_DIR, db_name)
        conf.set(phylogeny_path=tree_dir)
        conf.save()

        print("Phylogeny set to %s." % tree_dir)

    #
    # Make mashtree for database.
    #
    elif command == "mashtree":
        from expam.run import JSONConfig

        conf = JSONConfig(url=CONF_DIR)
        do_mashtree(conf, PHY_DIR, group)

    #
    # Create reads from some reference sequence. Used for unit tests.
    #
    elif command == "make_reads":
        from expam.sequences import make_reads

        make_reads(
            in_url=directory,
            l=length,
            n=n,
            out_url=out_url,
            e=0.0 if error_rate is None else float(error_rate)
        )

    #
    # Posthumously employ cutoff.
    #
    elif command == "cutoff":
        from expam.sequences import ls

        if os.path.exists(results_name):
            results_dir = results_name
        else:
            results_dir = os.path.join(DB_DIR, "results", results_name)

        tax_results_dir = os.path.join(results_dir, "tax")
        raw_results_dir = os.path.join(tax_results_dir, "raw")

        if not os.path.exists(results_dir):
            die("Could not find results %s!" % results_dir)

        file_names = [
            file_dir
            for file_dir in ls(tax_results_dir)
        ]

        for summary_file in file_names:
            print("Employing cutoff on sample %s..." % summary_file)

            with open(os.path.join(tax_results_dir, summary_file), "r") as f:
                data = [line.strip().split('\t') for line in f]

            total_classifications = sum(int(line[3]) for line in data[1:])
            total_splits = sum(int(line[6]) for line in data[1:])
            total_reads = total_classifications + total_splits

            if cpm is not None:
                cutoff = cpm * (total_reads / 1e6)
            elif cutoff is None:
                raise Exception("Require a cutoff value!")

            def valid_line(line):
                if int(line[2]) >= cutoff or int(line[5]) >= cutoff:
                    return True

            valid_taxids = [line[0] for line in data[1:] if valid_line(line)]
            valid_data = [line for i, line in enumerate(data) if i == 0 or valid_line(line)]

            # Overwrite summary.
            with open(os.path.join(tax_results_dir, summary_file), "w") as f:
                f.write("\n".join("\t".join(line) for line in valid_data))

            # Overwrite raw output.
            with open(os.path.join(raw_results_dir, summary_file), "r") as f:
                valid_data = []

                for line in f:
                    current_data = line.strip().split('\t')
                    if current_data[2] in valid_taxids:
                        valid_data.append('\t'.join(current_data))

            with open(os.path.join(raw_results_dir, summary_file), "w") as f:
                f.write('\n'.join(valid_data))

    #
    # Generate evenly-balanced phylogeny for sequences contained
    #   in directory.
    #
    elif command == "fake_phylogeny":
        from expam.run import JSONConfig
        from expam.sequences import format_name
        from expam.tree import simulate_balanced_phylogeny

        PHY_NAME = "expam_outtree.nwk"
        PHY_DIR = os.path.join("phylogeny", PHY_NAME)

        # Load configuration file.
        config = load_configuration_file(DB_DIR, return_config=True)

        # Get list of sequences.
        seq_names = [
            format_name(os.path.basename(url))
            for url in config.get_paths()
        ]

        if not seq_names:
            die("No sequences added to the database!")

        # Get valid save_directory.
        save_dir = os.path.join(DB_DIR, PHY_DIR)

        # Compute balanced phylogeny.
        newick_str = simulate_balanced_phylogeny(seq_names)

        with open(save_dir, "w") as f:
            f.write(newick_str)

        print("Newick file generated at %s." % save_dir)

        # Add phylogeny to expam configuration file.
        conf = JSONConfig(url=CONF_DIR)
        conf.set(phylogeny_path=PHY_DIR)
        conf.save()

        print("Database phylogeny set to %s." % save_dir)

    #
    # Count k-mers in sequence file.
    #
    elif command == "count":
        import numpy as np
        import pandas as pd

        from expam.c.extract import extract_kmers_from_read as get_kmers
        from expam.sequences import format_name
        from expam.stores import SequenceStore

        store = SequenceStore(k=k, out_dir="")

        seq_name = format_name(directory)
        store.add_sequence(directory)

        kmers = get_kmers(store[seq_name], k)
        values, counts = np.unique(kmers, return_counts=True)

        df = pd.DataFrame({"kmers": values, "counts": counts})
        df.to_csv(out_url, sep="\t", index=False, header=False)

    #
    # Map Jellyfish counts to expam.
    #
    elif command == "from_jellyfish":
        import numpy as np
        import pandas as pd

        from expam.c.extract import map_kmers
        from expam.run import calculate_n_chunks, KEYS_DATA_TYPE

        df = pd.read_csv(directory, sep=" ", names=["kmers", "counts"])
        seq = ("N".join(df['kmers'])).encode("utf8")

        kmers = np.zeros((len(df['kmers']), calculate_n_chunks(k) + 1), dtype=KEYS_DATA_TYPE)
        kmers[:, 0] = np.arange(len(df['kmers']), dtype=KEYS_DATA_TYPE) * (k + 1)
        map_kmers(seq, k, kmers)

        for i in range(calculate_n_chunks(k)):
            df['kmers_%d' % i] = kmers[:, i + 1]

        df.drop(columns=['kmers'], inplace=True)
        df.to_csv(out_url, sep="\t", index=False, header=False,
                  columns=['kmers_%d' % i for i in range(calculate_n_chunks(k))] + ['counts'])

    #
    # Plot time taken for tasks.
    #
    elif command == "plot":
        from expam.run import plot_data

        logs_dir = os.path.join(DB_DIR, "logs")
        _, t0 = get_t0(logs_dir)

        plot_data(logs_dir, t0)

    #
    #   Plot memory used by each process over time.
    #
    elif command == "plot_memory":
        from expam.sequences import ls

        def _get_mem(data):
            start, end = data.find("(") + 1, data.find(")")

            if start == -1 or end == -1:
                raise ValueError("Invalid data line %s!" % data)

            rss_string, vsz_string = data[start:end].split(", ")

            for point in (rss_string, vsz_string):
                mem_type, mem_used = point.split(" ")

                yield float(mem_used.replace("Gb", ""))

        def _r_yield_dict(obj, current_key=""):
            if isinstance(obj, dict):
                for key, value in obj.items():
                    yield from _r_yield_dict(value, "%s_%s" % (current_key, str(key)))

            else:
                yield current_key, obj

        job_types = {
            "_extract.log": "extract",
            "_union.log": "union",
        }
        plot_colours = {
            "extract": {
                "rss": "blue",
                "vsz": "black",
            },
            "union": {
                "rss": "green",
                "vsz": "purple"
            }
        }
        job_labels = {
            "extract": "extract",
            "union": "union",
        }

        logs_dir = os.path.join(DB_DIR, "logs")

        log_urls = [
            os.path.join(logs_dir, log_name)
            for log_name in ls(logs_dir, ext=".log")
            if "_main" not in log_name
        ]

        _, t0 = get_t0(logs_dir)

        # Collect total memory usage as well.
        cumulative = {
            "union": {
                "rss": 0,
                "vsz": 0,
            },
            "extract": {
                "rss": 0,
                "vsz": 0,
            }
        }

        fig, (memory_time_ax, cumulative_memory_ax) = plt.subplots(1, 2, figsize=(20, 10))

        for log_url in log_urls:

            with open(log_url, "r") as f:
                log_data = f.readlines()

            t, rss, vsz = [], [], []

            for line in log_data:
                if "of RAM" in line:
                    rss_mem, vsz_mem = _get_mem(line)
                    time_point = (get_time(line) - t0).total_seconds()

                    rss.append(rss_mem)
                    vsz.append(vsz_mem)
                    t.append(time_point / 3600)  # Time in hours.

            for mem_type, mem_data in (("rss", rss), ("vsz", vsz)):
                for k, v in job_types.items():
                    if k in log_url:
                        job_type = v

                        break
                else:
                    raise TypeError("Unknown process log type %s!" % os.path.basename(log_url))

                plot_label = job_labels[job_type]
                # Don't allow duplicate labels.
                job_labels[job_type] = ""

                memory_time_ax.plot(
                    t,
                    mem_data,
                    label=plot_label,
                    marker=".",
                    color=plot_colours[job_type][mem_type]
                )

                if mem_data:
                    # Add to cumulative memory usage.
                    cumulative[job_type][mem_type] += mem_data[-1]

        # Memory-over-time axis labels.
        memory_time_ax.set(
            xlabel="Time since start [hrs]",
            ylabel="Memory used [Gb]",
            title="Memory used by child workers"
        )
        memory_time_ax.legend(loc="best")

        # Bar chart of cumulative memory usage.
        labels = ["union", "extract"]
        rss = [cumulative["union"]["rss"], cumulative["extract"]["rss"]]
        vsz = [cumulative["union"]["vsz"], cumulative["extract"]["vsz"]]

        # Set axis positions.
        x = list(range(len(labels)))
        width = 0.35

        rss_x = list(v - width / 2 for v in x)
        vsz_x = list(v + width / 2 for v in x)

        # Plot bar charts.
        rss_bars = cumulative_memory_ax.bar(rss_x, rss, width, label="RSS")
        vsz_bars = cumulative_memory_ax.bar(vsz_x, vsz, width, label="VSZ")

        # Cumulative memory usage axis labels.
        cumulative_memory_ax.set(
            ylabel="Total memory used [Gb]",
            xticks=x,
            xticklabels=labels,
            title="Total memory used by workers",
        )
        cumulative_memory_ax.legend(loc="best")

        # Save figure to disk.
        fig.savefig(os.path.join(logs_dir, "memory_usage.png"))

    #
    #   Count how many genomes were processed in a given run.
    #
    elif command == "how_many":
        from expam.sequences import ls

        def _get_log_lines(log_dir):
            with open(log_dir, "r") as f:
                lines = f.readlines()

            return lines

        logs_dir = os.path.join(DB_DIR, "logs")
        log_files = [
            os.path.join(logs_dir, log_name)
            for log_name in ls(logs_dir, ext=".log")
        ]

        # First get extractions done.
        genomes_extracted = 0

        for log_file in log_files:
            if "_extract.log" in log_file:
                log_lines = _get_log_lines(log_file)

                for line in log_lines:
                    if "kmers sent!" in line:
                        genomes_extracted += 1

                    # Only count those during extraction phase.
                    elif "now at phase map!" in line:
                        break

        print("Genome k-mers extracted:\t%d." % genomes_extracted)

        # Now count how many unions done.
        genomes_unioned = []

        for log_file in log_files:
            if "_union.log" in log_file:
                log_lines = _get_log_lines(log_file)

                log_unions = 0
                for line in log_lines:
                    if "Importing" in line:
                        log_unions += 1

                genomes_unioned.append(log_unions)

        print("Genomes undergone union:\t%d." % min(genomes_unioned))
        print("\t Individual unions ~ {%s}" % " ".join(str(union) for union in genomes_unioned))

    else:
        die("Invalid command: " + str(command) + ".")


if __name__ == "__main__":
    main()



#cython: infer_types=True
#cython: boundscheck=False
#cython: wraparound=False
#cython: nonecheck=False
#cython: cdivision=True

cimport cython
from libc.math cimport ceil, floor
import numpy as np
cimport numpy as np

ctypedef np.uint8_t uint8_t
ctypedef np.uint16_t uint16_t
ctypedef np.uint64_t uint64_t
ctypedef np.int64_t int64_t


"""
Binary search implementation.

"""

cpdef binarysearch(uint64_t[:] target, uint64_t[:, :] kmers):
    """Binary search for target within array of kmers. Will return closest
    element in list if target is not found.

    :param target: Element to search for.
    :param kmers: Array to be searched within.
    :return: Whether element was found, and the corresponding index.
    :rtype: (int, int)

    """
    cdef:
        long long int i=0
        long long int L=0, U=kmers.shape[0]-1
        int j=0, W=kmers.shape[1]

    while True:
        if L > U:
            # Return the index with a fail signal.
            return 0, i
        i = <long long int>floor((U + L) / 2)
        for j in range(W):
            if kmers[i, j] > target[j]:
                U = i - 1
                break
            elif kmers[i, j] < target[j]:
                L = i + 1
                break
        else:
            return 1, i


def binarysearch_put(uint64_t[:, :] kmers, uint64_t[:, :] keys, uint16_t[:] values, uint16_t[:, :] lca_matrix, int new_id):
    """For all the elements in kmers, each coming from a node in the phylogeny
    with ID new_id, update their LCA values to incorporate this new_id. All
    possible LCA values have been computed in the LCA matrix.

    This means that every value in kmers should be in keys, as a binary search
    is executed to find the value of each kmer in keys.

    :param kmers: New items to be mapped.
    :param keys: (Sorted) Database keys to be searched against.
    :param values: Current LCA values for keys.
    :param lca_matrix: LCAs for each pair of nodes in the phylogeny.
    :param int new_id: ID of node containing kmers.
    :return: None (in-place)

    """
    cdef:
        long long int i, ind, M=kmers.shape[0]
        bint isin
        uint16_t current_id

    for i in range(M):
        # Find value in list.
        isin, ind = binarysearch(
            kmers[i, :],
            keys
        )

        # Find LCA of new and current value.
        current_id = values[ind]
        if current_id == new_id:
            pass
        elif current_id == 0:
            values[ind] = new_id
        elif current_id < new_id:
            values[ind] = lca_matrix[new_id, current_id]
        elif current_id > new_id:
            values[ind] = lca_matrix[current_id, new_id]


cpdef binarysearch_get(uint64_t[:, :] kmers, uint64_t[:, :] keys, uint16_t[:] values, uint16_t nullvalue):
    """Get the value of each element in kmers from keys. If the element is not
    found, nullvalue is used instead.

    :param kmers: Array of kmers to be searched for in keys.
    :param keys: Sorted database keys.
    :param values: Corresponding LCA values for keys.
    :param nullvalue: Value to be used if searched item is not found.
    :return: Array of values.
    :rtype: np.ndarray

    """
    cdef:
        long long int i, j, ind
        bint isin
        long long int M=kmers.shape[0]

    results = np.empty(M, dtype=np.uint16)
    cdef uint16_t[:] results_view = results
    for i in range(M):
        isin, ind = binarysearch(
            kmers[i, :],
            keys
        )
        if isin == 0:
            results_view[i] = nullvalue
        else:
            results_view[i] = values[ind]
    return results


"""
Insertion sort implementation.

"""

cdef unique(uint16_t[:] nodes, uint16_t[:] counts):
    insertion_sort(nodes, counts)
    nodes, counts = remove_duplicates(nodes, counts)
    return nodes, counts

cdef insertion_sort(uint16_t[:] nodes, uint16_t[:] counts):
    cdef int i = 1, j

    while i < nodes.shape[0]:
        j = i
        while j > 0 and nodes[j - 1] > nodes[j]:
            swap(nodes, j - 1, j)
            swap(counts, j - 1, j)
            j -= 1

        i += 1

cdef swap(uint16_t[:] arr, int i, int j):
    cdef uint16_t temp = arr[i]

    arr[i] = arr[j]
    arr[j] = temp

cdef remove_duplicates(uint16_t[:] nodes, uint16_t[:] counts):
    cdef int i = 0, j = 1

    while j < nodes.shape[0]:
        if nodes[i] == nodes[j]:
            counts[i] += counts[j]

        else:
            i += 1
            nodes[i] = nodes[j]
            counts[i] = counts[j]

        j += 1

    return nodes[:i + 1], counts[:i + 1]


"""
Classification routines.

"""

cpdef classify(uint64_t[:, :] kmers, uint64_t[:, :] keys, uint16_t[:] values, uint16_t[:, :] lca_matrix,
               uint16_t null_value, float alpha):
    cdef:
        uint16_t[:] nodes, counts
        uint16_t[:] raw_mapped_nodes
        uint16_t cls
        int candidates, cls_code = 0         # Innocent until proven guilty for class code.
        int not_null, required_counts, terminals
        str comp_string = ""

    # Map kmers using database.
    raw_mapped_nodes = binarysearch_get(kmers[:, 1:], keys, values, null_value)
    not_null = remove_null(raw_mapped_nodes, null_value)

    # Catch case of all unclassified.
    if not_null == 0:
        return -1, null_value, comp_string

    elif not_null < raw_mapped_nodes.shape[0]:
        cls_code = 1

    raw_mapped_nodes = raw_mapped_nodes[:not_null]

    # Convert mapped read to compressed string format.
    nodes, counts = kmers_to_compressed_form(raw_mapped_nodes)
    comp_string = compressed_to_string(nodes, counts)

    # Classification from lineage(s), with cutoff from alpha.
    required_counts = <int>ceil(alpha * kmers.shape[0])
    nodes, counts = unique(nodes, counts)

    # Get list of terminals.
    terminals = accumulate_candidates(nodes, counts, lca_matrix)

    # Find LCA of valid candidates.
    candidates, cls = choose_lca_candidate(nodes, counts, lca_matrix, terminals, required_counts)

    if candidates == 0:
        cls_code = 1
        cls = 1

    elif candidates > 1:
        cls_code = 1

    return cls_code, cls, comp_string


cdef int remove_null(uint16_t[:] raw_mapped_nodes, uint16_t null_value):
    cdef int j = 0, n = raw_mapped_nodes.shape[0], i

    for i in range(n):
        if raw_mapped_nodes[i] != null_value:
            raw_mapped_nodes[j] = raw_mapped_nodes[i]
            j += 1

    return j


cdef kmers_to_compressed_form(uint16_t[:] raw_mapped_nodes):
    """
    Return two arrays:
        1. Array of phylogeny nodes that consecutive kmers mapped to.
        2. The number of consecutive kmers that mapped to this node.

    """
    cdef:
        uint16_t[:] nodes, counts
        int n_kmers = raw_mapped_nodes.shape[0]
        int i, j = 0

    nodes = np.ndarray(n_kmers, dtype=np.uint16)
    counts = np.ones(n_kmers, dtype=np.uint16)
    
    nodes[0] = raw_mapped_nodes[0]

    for i in range(1, n_kmers):
        if raw_mapped_nodes[i] == nodes[j]:
            counts[j] += 1
        else:
            j += 1
            nodes[j] = raw_mapped_nodes[i]

    return nodes[:j + 1], counts[:j + 1]


def compressed_to_string(nodes, counts):
    return " ".join("p%d:%d" % (nodes[i], counts[i]) for i in range(nodes.shape[0]))


cdef int accumulate_candidates(uint16_t[:] nodes, uint16_t[:] counts, uint16_t[:, :] lca_matrix):
    cdef:
        int i, terminals = 0, n_nodes = nodes.shape[0]
        bint new_terminal

    for i in range(n_nodes - 1, -1, -1):
        new_terminal = check_terminals(nodes, counts, lca_matrix, i, terminals)

        if new_terminal:
            terminals = add_terminal(nodes, counts, i, terminals)

    return terminals


cdef bint check_terminals(uint16_t[:] nodes, uint16_t[:] counts, uint16_t[:, :] lca_matrix, int i, int terminals):
    cdef:
        int j, n = nodes.shape[0] - 1

    for j in range(n, n - terminals, -1):
        if lca_matrix[nodes[j], nodes[i]] == nodes[i]:          # Remember that j > i (Indexing LT lca matrix).
            counts[j] += counts[i]
            return 0

    return 1


cdef int add_terminal(uint16_t[:] nodes, uint16_t[:] counts, int i, int terminals):
    cdef:
        int minus_t = (nodes.shape[0] - 1) - terminals

    nodes[minus_t] = nodes[i]
    counts[minus_t] = counts[i]

    return terminals + 1


cdef choose_lca_candidate(uint16_t[:] nodes, uint16_t[:] counts, uint16_t[:, :] lca_matrix, int terminals,
                                   int required_counts):
    cdef:
        int n = nodes.shape[0] - 1, start
        int candidates = 0
        uint16_t lca = 1

    # Base case.
    start = n
    while start > n - terminals:
        if counts[start] >= required_counts:
            candidates = 1
            lca = nodes[start]

            break

        start -= 1

    # Take LCA of remaining valid candidates.
    for i in range(start - 1, n - terminals, -1):
        if counts[i] >= required_counts:
            candidates += 1

            if lca < nodes[i]:
                lca = lca_matrix[nodes[i], lca]
            else:   # Note that lca is never equal to any terminal in `nodes`.
                lca = lca_matrix[lca, nodes[i]]

    return candidates, lca



import io
import gzip
import os
import random
from os import listdir
from os.path import basename, isfile, join

from expam.c.extract import reverse_complement_combine
from expam.run import timeit

# Parsing of compressed files.
DEFAULT_MODE = "rb"
DEFAULT_OPENER = open
COMP_PARSE = {
    ".tar.gz": {"mode": "rb", "opener": gzip.open},
    ".gz": {"mode": "rb", "opener": gzip.open}
}


def get_opener(file_name):
    for suffix in COMP_PARSE:

        if check_suffix(file_name, suffix):
            file_name = file_name.replace(suffix, '')
            mode, opener = COMP_PARSE[suffix]["mode"], COMP_PARSE[suffix]["opener"]

            break

    else:
        mode, opener = DEFAULT_MODE, DEFAULT_OPENER

    return file_name, (mode, opener)


def check_suffix(string, sfx):
    string = string.strip()

    if string[-len(sfx):] == sfx:
        return True
    return False


def ls(url, ext=""):
    null_ext = not len(ext)

    if isfile(url) and (url[-len(ext):] == ext or null_ext):
        return [url]
    else:
        return [
            f for f in listdir(url)
            if isfile(join(url, f)) and (f[-len(ext):] == ext or null_ext)
        ]


def format_name(name, remove_comp=False):
    name = basename(name)
    # Must not be a digit. If it is, add 'ref' to the start.
    if name.isdigit():
        name = "ref" + name

    # Extract sequence name.
    if ".fna" in name:
        name = name.replace(".fna", "")
    elif ".faa" in name:
        name = name.replace(".faa", "")

    if remove_comp:
        for suffix in COMP_PARSE:
            if check_suffix(name, suffix):
                name = name.replace(suffix, '')

                break

    return name


def extract_id(metadata):
    # Check for kraken format.
    if "kraken" in metadata:
        start = metadata.rfind("|") + 1
    else:
        # Standard fasta format.
        if ">" in metadata:
            start = metadata.find(">") + 1
        elif "@" in metadata:
            start = metadata.find("@") + 1

    end = metadata.find(" ")
    if not end:  # Case where no metadata is provided.
        end = len(metadata)

    # Contig name, including potential contig version number.
    contig_id = metadata[start:end]
    # Concatenate multiple contig versions.
    version_start = contig_id.rfind(".")
    if version_start:
        contig_id = contig_id[:version_start]
    return contig_id


@timeit
def iterate_sequence(url):
    with open(url, "r") as f:
        seqID, contig = "", ""
        for sequence in f:
            # Strip newlines.
            if '\n' in sequence:
                sequence = sequence.strip('\n')
            if ">" in sequence:
                if contig != "":
                    # Return the previous sequence, if there is one.
                    yield (seqID, contig)
                # Find seqID.
                if "|" in sequence:
                    # >x|seqID|... format.
                    seqID = sequence.split("|")[1]
                else:
                    # >seqID ... format.
                    seqID = sequence.split(" ")[0]
            else:
                # Combine sequence.
                contig += sequence
        # End case: yeild last contig.
        if contig != "":
            yield seqID, contig


@timeit
def get_sequence(url):
    """
    Extract the sequence from FASTA source file @ url.

    :param url: String.
    :return: String.
    """
    sequence = io.BytesIO()
    _, (mode, opener) = get_opener(url)

    with opener(url, mode) as f:
        for line in f:
            line = line.strip()
            # Skip information line at start of FASTA format.
            if b">" in line:
                sequence.write(b"N")

            else:
                sequence.write(line)

    return sequence.getvalue()


def make_reads(in_url, l, n, out_url, e=0.0):
    """
    Make n reads from sequence at url, from a expam maker class.
    """
    maker = Reads(l, out_url)
    maker.make(in_url, n, error=e)


class Reads:
    def __init__(self, l, out_dir):
        # Length of reads.
        self.l = l
        # Directory to save reads.
        self.out_dir = out_dir

    def make(self, f_url, n_reads, error=0.0):
        """
        Produce a number of FASTQ reads from a sequence.

        :param error: Float - Rate that errors will be put into reads (error ~ reads with errors / reads).
        :param f_url: String - Url of the sequence.
        :param n_reads: Int - number of reads to produce.
        """
        if error > 1:
            if error > 100:
                raise ValueError("Error rate must be either < 1 or < 100!")

            error /= 100

        nt_codes = {0: b"A", 1: b"C", 2: b"G", 3: b"T"}
        bad_quality = bytes("N" * 75, encoding="utf8")

        # Check formatting of sequence.
        self.check_fna(f_url)
        # Extract sequence string from f_url.
        sequence = get_sequence(f_url)

        # Extract the file name from the path.
        forward_name, reverse_name = self.get_file_name(f_url)

        # Generate the reads.
        i = 0
        forward_read_stream = io.BytesIO()
        reverse_read_stream = io.BytesIO()

        while i < n_reads:

            # Generate a start index for the read.
            index = random.randint(0, len(sequence) - self.l)
            # Get read.
            read = sequence[index: index + self.l]

            # Check quality.
            if bad_quality in read:
                # This is deemed poor quality. Try again.
                continue

            rid = ("R" + str(abs(hash(str(index))))).encode('utf8')
            meta_line = b">%b Read from %b.\n" % (rid, f_url.encode('utf8'))

            # Generate error.
            has_error = random.randint(0, 100) / 100 < error

            if has_error:
                sub_location = random.randint(0, len(read) - 1)
                nt = nt_codes[random.randint(0, 3)]

                read_string = b"%b%b%b" % (read[:sub_location], nt, read[sub_location + 1:])

            else:
                read_string = b"%b" % read

            forward_reverse_string = reverse_complement_combine(read_string, read_string)
            l = len(read_string)

            forward_string = forward_reverse_string[:l]
            reverse_string = forward_reverse_string[-l:]

            forward_read_stream.write(meta_line)
            forward_read_stream.write(forward_string)
            forward_read_stream.write(b'\n')

            reverse_read_stream.write(meta_line)
            reverse_read_stream.write(reverse_string)
            reverse_read_stream.write(b'\n')

            i += 1

        # Write to file.
        with open(os.path.join(self.out_dir, forward_name), "wb") as f:
            f.write(forward_read_stream.getvalue())

        forward_read_stream.seek(0)
        forward_read_stream.truncate()

        with open(os.path.join(self.out_dir, reverse_name), "wb") as f:
            f.write(reverse_read_stream.getvalue())

        reverse_read_stream.seek(0)
        reverse_read_stream.truncate()

    @staticmethod
    def check_fna(url):
        if not ".fna" in url:
            raise TypeError("Sequence type must be .fna.")

    @staticmethod
    def get_file_name(url):
        path = os.path.normpath(url)
        path = path.split(os.sep)

        forward_name = path[-1].rstrip(".fna") + "_1.fa"
        reverse_name = path[-1].rstrip(".fna") + "_2.fa"

        return forward_name, reverse_name



import random
from math import floor
import json
import os
import traceback

import numpy as np
import pandas as pd

PHYLA_COLOURS = [("Proteobacteria", "#E85343"), ("Bacteroidetes", "#30B674"),
                 ("Firmicutes", "#6585C3"), ("Actinobacteria", "#FEDB45")]


def simulate_balanced_phylogeny(names):
    """
    :param names: List - list of leaf names.
    """

    # Declare an index.
    index = Index()

    # Insert all children.
    for name in names:
        # Make child node.
        childNode = Location(
            name=name,
            type="Leaf",
            dist=1.0
        )
        # Insert into tree.
        index.append(childNode)

    # Naming string.
    _NAME = len(names) - 1

    # Repeatedly join pairs until a full tree is made.
    while len(names) > 1:
        new_names = []

        while len(names) > 0:
            try:
                # Get next two children.
                child_names = []
                i = 0
                while i < 2:
                    child_names.append(names.pop())
                    i += 1

            except IndexError:
                # Only one child left. Just pass on the name.
                new_names.append(child_names[0])
                break

            # Join these children:
            # Make a parent node.
            parentName = str(_NAME)
            _NAME -= 1
            # Make a new Location.
            parentNode = Location(
                name=parentName,
                type="Branch"
            )
            print(f'New Parent {parentName}...')
            # Apply join.
            index.join(
                child_names[0],
                child_names[1],
                parentNode
            )

            # Append parent name to new_names.
            new_names.append(parentName)

        # New lot of children.
        names = new_names

    # Convert tree to newick format.
    return index.to_newick() + ";"


def propose_lca(c1, c2):
    """
    Given two binary coordinates, compute the shortest common prefix.

    :param coord: Indexable.
    :return: List.
    """
    m = min(len(c1), len(c2))
    if m == 0:
        return []

    for i in range(1, m + 1):
        if c1[-i] != c2[-i]:
            if i == 1:
                return []
            else:
                return c1[-i + 1:]
    else:
        return c1[-m:]


class Location:
    """Represents a node in the phylogeny."""
    is_root = False

    def __init__(self, name="", type="", dist=0.0, coord=None, accession_id=None, taxid=None, **kwargs):
        """__init__ 

        :param name: name of node, defaults to ""
        :type name: str, optional
        :param type: Leaf or Branch, defaults to ""
        :type type: str, optional
        :param dist: distance to parent node, defaults to 0.0
        :type dist: float, optional
        :param coord: binary coordinate from root to node, defaults to None
        :type coord: list, optional
        :param accession_id: NCBI accession id, defaults to None
        :type accession_id: str, optional
        :param taxid: NCBI taxonomy id, defaults to None
        :type taxid: int, optional
        :ivar name: node name
        :ivar type: "Leaf" or "Branch"
        :ivar distance: distance to parent node
        :ivar coordinate: list of binary binary numbers representing path from root to node
        :ivar nchildren: number of children below this node
        :ivar accession_id: NCBI accession id (only valid for leaves)
        :ivar taxid: NCBI taxonomy id
        """
        self._name = name
        self._type = type
        self._distance = dist
        self._coordinate = [] if coord is None else coord
        self._nchildren = 0
        self._accession_id = accession_id
        self._taxid = taxid

        self.children = []
        self.primaryChild = None

    def __str__(self):
        return (f"Location - {self._name} ({self._type}):"
                f"\n\t+ Coordinate: {self._coordinate}"
                f"\n\t+ Children: {self._nchildren}")

    def name():
        doc = "The name property."

        def fget(self):
            return self._name

        def fset(self, value):
            self._name = value

        def fdel(self):
            del self._name

        return locals()

    name = property(**name())

    def type():
        doc = "The type property."

        def fget(self):
            return self._type

        def fset(self, value):
            self._type = value

        def fdel(self):
            del self._type

        return locals()

    type = property(**type())

    def distance():
        doc = "The distance property."

        def fget(self):
            return self._distance

        def fset(self, value):
            self._distance = value

        def fdel(self):
            del self._distance

        return locals()

    distance = property(**distance())

    def coordinate():
        doc = "The coordinate property."

        def fget(self):
            return self._coordinate

        def fset(self, value):
            self._coordinate = value

        def fdel(self):
            del self._coordinate

        return locals()

    coordinate = property(**coordinate())

    def nchildren():
        doc = "The nchildren property."

        def fget(self):
            return self._nchildren

        def fset(self, value):
            self._nchildren = value

        def fdel(self):
            del self._nchildren

        return locals()

    nchildren = property(**nchildren())

    def accession_id():
        doc = "The accession id property."

        def fget(self):
            return self._accession_id

        def fset(self, value):
            self._accession_id = value

        def fdel(self):
            del self._accession_id

        return locals()

    accession_id = property(**accession_id())

    def taxid():
        doc = "The accession id property."

        def fget(self):
            return self._taxid

        def fset(self, value):
            self._taxid = value

        def fdel(self):
            del self._taxid

        return locals()

    taxid = property(**taxid())

    @classmethod
    def make_branch(cls, branch_name):
        return Location(branch_name, "Branch", 0)

    @classmethod
    def make_leaf(cls, tree, leaf_name, dist):
        return Location(leaf_name, "Leaf", dist)

    def add_child(self, child, primary_child=False):
        self.children.__setitem__(child.name, child)

        if primary_child:
            self.set_primary(child)

    def set_primary(self, node):
        self.primaryChild = node

    def save(self, out_dir):
        fname = f'{self._name}.loc'
        with open(os.path.join(out_dir, "phylogeny", "loc", fname), 'w') as f:
            json.dump(self.__dict__, f)


class Index:
    """ Phylogeny index that can load, save and manipulate Newick trees.
    """
    def __init__(self):
        self._pointers = {}  # Pointers is a map from node name to pool location.
        self.pool = [Location()]  # Empty location to represent an 'unclassified' location.

    def __len__(self):
        return len(self.pool)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f'<Phylogeny Index, length={len(self)}>'

    @classmethod
    def load_newick(cls, path):
        """load_newick Load Newick tree from file.

        :param path: path to Newick file
        :type path: str
        :raises OSError: file does not exist
        :return: name of leaves and phylogeny Index object
        :rtype: list[str], expam.tree.Index
        """
        newick_str = ""

        if not os.path.exists(path):
            raise OSError("Couldn't find path %s." % path)

        with open(path, "r") as f:
            for line in f:
                newick_str += line.strip()

        return cls.from_newick(newick_str)

    @classmethod
    def from_newick(cls, newick_string):
        """from_newick Parse Newick string.

        :param newick_string: Newick string encoding tree.
        :type newick_string: str
        :return: name of leaves and phylogeny Index object
        :rtype: list[str], expam.tree.Index
        """

        # Remove whitespace.
        newick_string = newick_string.replace(" ", "")

        print("* Initialising node pool...")
        index_pool = cls.init_pool(newick_string)

        print("* Checking for polytomies...")
        cls.resolve_polytomies(index_pool)

        print("* Finalising index...")
        leaf_names, index = cls.from_pool(index_pool)

        return leaf_names, index

    @staticmethod
    def init_pool(newick):
        pool = [Location()]

        i = 0
        parent_stack = []
        currentNode = None

        #
        # Create a new node.
        #
        def new_node(name, type):
            nonlocal currentNode

            currentNode = Location(name, type)

            pool.append(currentNode)

            # Top node has no ancestor.
            if len(parent_stack) > 0:
                parent_stack[-1].children.append(currentNode)

        #
        # Get node name.
        #
        def close_branch():
            nonlocal parent_stack, currentNode, i, newick, NEWICK_PARSE

            current_branch = parent_stack.pop()

            # Get this branch's name, if it is given.
            if newick[i + 1] not in NEWICK_PARSE:
                i += 1
                branch_name = parse_string()

                current_branch.name = branch_name

            currentNode = current_branch

        #
        # New branch in phylogeny and parent stack.
        #
        def new_branch():
            nonlocal currentNode

            new_node(name="", type="Branch")
            parent_stack.append(currentNode)

        #
        # New Leaf in phylogeny.
        #
        def new_leaf():
            nonlocal currentNode

            new_node(name="", type="Leaf")

        #
        # Parse name.
        #
        def parse_string(force_digits=False):
            nonlocal i  # Inherit current index.

            j = i + 1

            while newick[j] not in NEWICK_PARSE:
                j += 1
            string = newick[i:j]

            # Check formatting.
            if force_digits:
                try:
                    string = float(string)

                except ValueError:
                    raise ValueError("Invalid distance declaration %s!" % newick[i:j])

            i = j - 1  # Update current index position.

            return string

        #
        # Clear current node (stack-ish).
        #
        def clear_node():
            nonlocal currentNode

            currentNode = None

        #
        # Set distance of current node.
        #
        def set_node_distance():
            nonlocal currentNode, i

            i += 1  # Skip ':' character.
            currentNode.distance = parse_string(force_digits=True)

        #
        # Do nothing.
        #
        def finisher():
            nonlocal i

            i = len(newick)

        # Parsing functions for Newick format.
        NEWICK_PARSE = {
            "(": new_branch,
            ")": close_branch,
            ",": clear_node,
            ":": set_node_distance,
            ";": finisher
        }

        # Parsing loop.
        while i < len(newick):
            if newick[i] in NEWICK_PARSE:
                NEWICK_PARSE[newick[i]]()

            else:
                new_leaf()

                currentNode.name = parse_string()

            i += 1

        # Return pool array for insertion into expam Index.
        return pool

    @staticmethod
    def resolve_polytomies(pool):
        """
        If the phylogeny contains polytomies, continually join the first two
        children with parents of distance 0 until the polytomy is resolved.

        :param pool: List.
        :return: None
        :rtype: None
        """
        # Processing loop.
        i = 1
        while i < len(pool):
            node = pool[i]

            # Continually join first two nodes.
            while len(node.children) > 2:
                print("\tPolytomy (degree=%d) detected! Resolving..."
                      % len(node.children))

                # Declare new parent of distance 0.
                newParent = Location(
                    name="",
                    type="Branch",
                    dist=0.0
                )

                # Replace children.
                for _ in range(2):
                    newParent.children.append(node.children.pop(0))

                # Insert into pool.
                node.children.append(newParent)
                pool.insert(i + 1, newParent)

            i += 1

    @classmethod
    def from_pool(cls, pool):
        leaf_names = []

        # Initialise expam Index.
        index = cls()
        index.pool = pool

        branch_id = 1
        for i, node in enumerate(pool):
            if i == 0:
                continue

            if node.type == "Branch":
                # Set branch name.
                node.name = str(branch_id)
                branch_id += 1

            else:
                # Store leaf names.
                leaf_names.append(node.name)

            # Set name mapping.
            index._pointers[node.name] = i

        # Replace children objects with respective names.
        # Note: This requires the name mapping to be complete...
        for node in pool[1:]:
            if node.type == "Branch":
                i = 0
                while i < len(node.children):
                    node.children[i] = node.children[i].name

                    i += 1

        # Set (binary) coordinates.
        index.update_coordinates()

        # Update nchildren.
        index.update_nchildren()

        return leaf_names, index

    def append(self, item):
        # Declare an index.
        index = len(self.pool)

        # Set pointer.
        self._pointers[item.name] = index

        self.pool.append(item)

    def join(self, a, b, glue, dist=None):
        if dist is None:
            dist = [1.0, 1.0]
        i, j = self._pointers[a], self._pointers[b]
        i, j = self.smaller(i, j)  # i < j.

        one, two = j, self.neighbouring(j)
        three = self.neighbouring(i)
        self.move([one, two], three)

        self.insert(i, glue)

        # Update number of children and distances.
        glue.children.extend((a, b))

        for i, child in enumerate([a, b]):
            glue.nchildren += 1 + ((self[child].type == "Branch") * self[child].nchildren)

            self[child].distance = dist[i]

    def move(self, i, j):
        # NOTE: i > j.
        width = i[1] - i[0]

        if i[0] == j:
            return

        # Delete affected pointers.
        for n in range(j, i[1]):
            del self._pointers[n]

        # Retrieve items to be moved.
        temp = []
        for n in range(width):
            temp.append(self.pool.pop(i[1] - (n + 1)))

        # Insert these items in at j. Since they were popped in reverse order,
        # insert append them in the same index to revert back to the original
        # order.
        for obj in temp:
            self.pool.insert(j, obj)

        # Update pointers.
        for n in range(j, i[1]):
            self._pointers[n] = self.pool[n].name

    def insert(self, i, item):
        # Update pointers.
        for n in range(i, len(self.pool)):
            loc = self.pool[n]

            self._pointers[loc.name] = n + 1

        self.pool.insert(i, item)
        self._pointers[item.name] = i

    @staticmethod
    def _format_node_name(node_name):
        return str(node_name).lstrip('p')

    def __setitem__(self, key, value):
        key = self._format_node_name(key)  # Delete node identifier from results file.
        pool_index = self._pointers[key]
        self.pool[pool_index] = value

    def __getitem__(self, key):
        key = self._format_node_name(key)
        pool_index = self._pointers[key]
        return self.pool[pool_index]

    def __delitem__(self, key):
        self.pool.pop(self._pointers[str(key)])

    def neighbouring(self, i):
        return i + 1 + self.pool[i].nchildren

    def update_coordinates(self):
        coord, i = [-1], 1

        while i < len(self.pool):
            while coord[0] + 1 > 1:
                coord.pop(0)

            coord[0] += 1

            if self.pool[i].type == "Branch":
                self.pool[i].coordinate = list(coord[:-1])

                coord.insert(0, -1)
            else:  # It's a leaf.
                self.pool[i].coordinate = list(coord[:-1])

            i += 1

    def update_nchildren(self):
        for j in range(len(self.pool) - 1, 0, -1):
            node = self.pool[j]

            node.nchildren = len(node.children) + sum([
                self[child].nchildren
                for child in node.children
            ])

    def coord(self, coordinate):
        """coord Return Location (node) at coordinate.

        :param coordinate: binary list representing path to node
        :type coordinate: list
        :return: node in tree
        :rtype: expam.tree.Location
        """
        pool_index = 1

        for step in coordinate[::-1]:
            pool_index += 1
            pool_index += int(step) * (self.pool[pool_index].nchildren + 1)

        return self.pool[pool_index]

    def print_index(self):
        for l in self.pool:
            print(l.name, l.coordinate, l.nchildren)

    @staticmethod
    def smaller(a, b):
        if a < b:
            return a, b
        return b, a

    def to_newick(self):
        """to_newick Output tree to Newick format.

        :return: Newick format tree
        :rtype: str
        """
        NODE_FORMAT = "$node$"

        newick = "(%s,%s)%s;" % (NODE_FORMAT, NODE_FORMAT, self.pool[1].name)
        node_formats = {
            "Branch": "(%s,%s){name}:{distance}" % (NODE_FORMAT, NODE_FORMAT),
            "Leaf": "{name}:{distance}"
        }

        for node in self.pool[2:]:
            node_data = node_formats[node.type].format(name=node.name, distance=node.distance)
            newick = newick.replace(NODE_FORMAT, node_data, 1)

        return newick

    def yield_child_nodes(self, node_name):
        """yield_child_nodes Yields node and children nodes (both branches and leaves).

        :param node_name: name of node to start yielding from
        :type node_name: str
        :yield: node names at or below node_name
        :rtype: str
        """
        node_name = self._format_node_name(node_name)
        active_nodes = [node_name]

        while active_nodes:
            node = active_nodes.pop(0)

            yield node
            active_nodes.extend(self[node].children)

    def yield_leaves(self, node_name):
        """yield_leaves Yield only the leaves at or below some node.

        :param node_name: node to retrieve leaves from.
        :type node_name: str
        :yield: leaf names at or below node_name.
        :rtype: str
        """
        node_name = self._format_node_name(node_name)

        if self[node_name].type == "Leaf":
            yield node_name
            return

        branches = [node_name]

        while branches:
            branch = branches.pop(0)

            for child in self[branch].children:
                if self[child].type == "Leaf":
                    yield child
                else:
                    branches.append(child)

    def get_child_nodes(self, node_name):
        """get_child_nodes Return list of nodes at or below node_name.

        :param node_name: name of node 
        :type node_name: str
        :return: list of node names
        :rtype: list[str]
        """
        return list(self.yield_child_nodes(node_name))

    def get_child_leaves(self, node_name):
        """get_child_leaves Get list of leaves at or below node_name.

        :param node_name: name of node
        :type node_name: str
        :return: list of leaf names
        :rtype: list[str]
        """
        return list(self.yield_leaves(node_name))

    def draw_results(self, file_path, out_dir, skiprows=None, groups=None, cutoff=None, cpm=None, colour_list=None,
                     name_to_taxon=None, use_phyla=False, keep_zeros=True, use_node_names=True):
        counts = pd.read_csv(file_path, sep='\t', index_col=0, header=0, skiprows=skiprows)
        self.draw_tree(out_dir, counts=counts, groups=groups, cutoff=cutoff, cpm=cpm, colour_list=colour_list,
                       name_to_taxon=name_to_taxon, use_phyla=use_phyla, keep_zeros=keep_zeros,
                       use_node_names=use_node_names)

    def draw_tree(self, out_dir, counts=None, groups=None, cutoff=None, cpm=None, colour_list=None, name_to_taxon=None,
                  use_phyla=False, keep_zeros=True, use_node_names=True):
        try:
            import ete3.coretype.tree
            from ete3 import AttrFace, CircleFace, faces, Tree, TreeStyle, NodeStyle, TextFace

        except ModuleNotFoundError as e:
            print("Could not import ete3 plotting modules! Error raised:")
            print(traceback.format_exc())
            print("Skipping plotting...")

            return

        """
        Phylogenetic printing of nodes.
        """
        for node in self.pool:
            if node.type == 'Branch' and node.name[0] != 'p':
                node.name = 'p' + node.name

        """
        Create PhyloTree.
        """
        newick_string = self.to_newick()
        tree = Tree(newick_string, format=1)

        """
        Take care of groupings.
        """
        if groups is None:
            groups = [(None, (section,)) for section in counts.columns]

        # *** Relies on counts being defined.
        if counts is not None:
            for _, group in groups:
                if len(group) > 1:
                    group_name = group[0]

                    # Conglomerate counts within the group.
                    counts.loc[:, group_name] = counts[list(group)].sum(axis=1)
                    counts.drop(labels=list(groups[1:]), axis=1, inplace=True)

            sections = counts.columns.tolist()
            nodes = counts.index.tolist()

            """
            Employ cutoff.
            """
            if cutoff is None and cpm is None:
                nodes_with_counts = nodes

            else:
                nodes_with_counts = set()

                for section in sections:
                    total = sum(counts[section])
                    section_cutoff = max(cutoff, (total / 1e6) * cpm)

                    for index in nodes:
                        if index not in nodes_with_counts:
                            if np.any(counts.loc[index, :] >= section_cutoff):
                                nodes_with_counts.add(index)

                nodes_with_counts = list(nodes_with_counts)

            """
            Attempt pruning of tree.
            """
            try:
                if not keep_zeros:
                    tree.prune(nodes_with_counts, preserve_branch_length=True)

            except ete3.coretype.tree.TreeError:
                print("Tree pruning failed, aborting tree drawing!")
                return

            """
            Colour generation.
            """
            max_vector = tuple(counts.max())
            colours = [colour for colour, _ in groups]
            colour_generator = ColourGenerator(colours, max_vector, colour_list=colour_list)

        """
        Function to render any given node.
        """

        def layout(node):
            nonlocal name_to_taxon

            if node.is_leaf():
                if use_node_names:
                    faces.add_face_to_node(AttrFace('name', fsize=20), node, column=0, position='aligned')

                if use_phyla:
                    for phyla, colour in PHYLA_COLOURS:
                        lineage = name_to_taxon[node.name]

                        if phyla in lineage:
                            tax_face = TextFace("   ")
                            tax_face.background.color = colour
                            faces.add_face_to_node(tax_face, node, column=1, position='aligned')

            if (counts is not None) and (node.name in counts.index):
                node_counts = tuple(counts.loc[node.name, :])

                if any(node_counts):
                    colour, _ = colour_generator.generate(node_counts)

                    ns = NodeStyle()
                    ns['bgcolor'] = colour
                    node.set_style(ns)

        """
        Render tree.
        """
        ts = TreeStyle()
        ts.mode = "c"
        ts.show_leaf_name = False
        ts.layout_fn = layout
        ts.force_topology = False
        ts.allow_face_overlap = False
        ts.draw_guiding_lines = True
        ts.root_opening_factor = 1
        tree.render(
            out_dir,
            tree_style=ts,
            dpi=300
        )
        print("Phylogenetic tree written to %s!" % out_dir)


class HexColour:
    def __init__(self, hex_string):
        self.colour = hex_string.lstrip("#")

    def __add__(self, other):
        rgb_self = self.to_rgb()
        rgb_other = other.to_rgb()

        rgb_mix = rgb_self + rgb_other
        return rgb_mix.to_hex()

    def __iadd__(self, other):
        mix = self + other
        self.colour = mix.colour

    def __str__(self):
        return "#%s" % self.colour

    def to_rgb(self):
        hx = self.colour
        assert len(hx) == 6

        rgb = tuple(int(hx[2 * i:2 * (i + 1)], 16) for i in range(3))
        return RGBColour(rgb)

    def opaque(self, q):
        if q == 1:
            return self

        rgb = self.to_rgb().opaque(q)

        return rgb.to_hex()


class RGBColour:
    def __init__(self, tup):
        self.colour = tup

    def __add__(self, other):
        assert len(self.colour) == len(other.colour) == 3

        rgb = tuple(
            floor(0.5 * (self.colour[i] + other.colour[i]))
            for i in range(3)
        )
        return RGBColour(rgb)

    def __iadd__(self, other):
        mix = self + other
        self.colour = mix.colour

    @staticmethod
    def _format_rgb(num):
        f_hx = hex(num).lstrip('0x').upper()

        if len(f_hx) < 2:
            f_hx = ('0' * (2 - len(f_hx))) + f_hx

        return f_hx

    def to_hex(self):
        rgb = self.colour
        assert len(rgb) == 3

        hx = "#" + "".join(self._format_rgb(factor) for factor in rgb)
        return HexColour(hx)

    def opaque(self, q):
        return RGBColour(tuple(
            floor(255 - (q * (255 - v)))
            for v in self.colour
        ))


class RandomColour:
    def __init__(self, rng=None, seed=1):
        self.alpha = "0123456789ABCDEF"

        if rng is not None:
            if not isinstance(rng, random.Random):
                raise TypeError("rng must be instance of random.Random!")

            self.rng = rng
        else:
            if seed is not None:
                self.rng = random.Random(seed)
            else:
                self.rng = random.Random()

    def __next__(self):
        return "#" + "".join(self.rng.choices(self.alpha, k=6))


class ColourGenerator:
    def __init__(self, colours, max_vector, colour_list=None):
        """

        :param colours:
        :param max_vector:
        :param colour_list:
        """
        """
        Ensure we have enough colours for plotting.
        """
        g = RandomColour()

        # Ensure we have enough colours for each group.
        required_colours = sum(colour is None for colour in colours)
        colours = [colour for colour in colours if colour is not None]

        if required_colours > 0:
            if colour_list is None:
                colour_list = [next(g) for _ in range(required_colours)]

            else:
                colours_to_generate = required_colours - len(colour_list)
                colour_list.extend(
                    [next(g) for _ in range(colours_to_generate)]
                )

            colours.extend(colour_list)

        self.colours = [HexColour(colour) for colour in colours]

        #
        self.max_vector = max_vector

    def _calculate_colour(self, score_vector, scorer):
        colour = HexColour("#FFFFFF")  # White background.

        # Artificial scaling factor.
        non_zero_indices = [i for i, score in enumerate(score_vector) if score > 0]
        alpha = sum([score_vector[i] for i in non_zero_indices]) / sum([scorer[i] for i in non_zero_indices])

        for i, score in enumerate(score_vector):
            if score > 0:
                colour += self.colours[i].opaque(score / scorer[i])

        return str(colour), alpha

    def generate(self, vector):
        return self._calculate_colour(vector, self.max_vector)



import datetime
import json
import logging
import subprocess

import math
import os
import queue
import time
import sys
from ctypes import *
from multiprocessing import Array, Pipe, Queue, shared_memory, Value
from timeit import default_timer

import matplotlib.pyplot as plt
import expam
import numpy as np

KEYS_DATA_TYPE = np.uint64
KEYS_DATA_TYPE_SIZE = np.dtype(KEYS_DATA_TYPE).itemsize
KEYS_DTYPE_STR = "uint64"

MAP_DATA_TYPE = np.uint16
MAP_DATA_TYPE_SIZE = np.dtype(MAP_DATA_TYPE).itemsize
MAP_DTYPE_STR = "uint16"

CHUNK_SIZE = 32
UNION_RATIO = 4 / 5

TIMEOUT = 1e-4

ACCESSION_ID_PATH = os.path.join("phylogeny", "accession_ids.csv")
COORDS_PATH = os.path.join("phylogeny", "coords.json")
LCA_MATRIX_PATH = os.path.join("phylogeny", "lca_matrix.npy")
DB_PATH = os.path.join("database", "expam_db.h5")

NULL_VALUE = 0  # Return value for foreign kmer.
EXPAM_TEMP_EXT = "expam_temp"

LOG_FORMAT = logging.Formatter(
    fmt='%(asctime)s... %(message)s',
    datefmt='%m/%d/%Y %I:%M:%S %p'
)


def die(msg):
    print("ERROR\n", msg, "\n")
    sys.exit(1)


class Timer:
    def __init__(self):
        self.start = None
        self.elapsed = None

    def __enter__(self):
        self.start = default_timer()
        return self

    def __exit__(self, exit_type, value, traceback):
        self.elapsed = default_timer() - self.start

    def __str__(self):
        return self.verbose()

    def __float__(self):
        return self.elapsed

    def verbose(self):
        if self.elapsed is None:
            return "<not_measured>"
        return str(self.elapsed) + "s"


class JSONConfig:
    _parse_data_type = {
        "uint8": np.uint8,
        "uint16": np.uint16,
        "uint32": np.uint32,
        "uint64": np.uint64,
        "int8": np.int8,
        "int16": np.int16,
        "int32": np.int32,
        "int64": np.int64,
    }

    def __init__(self, url=None, name=""):
        self.url = url

        # Default values.
        self.params = {
            "name": name,
            "groups": {
                'default': {
                    'sequence_files': [],
                    'phylogeny_path': [],
                    'k': None,
                    's': None,
                },
            },
            "phylogeny_path": None,
            "k": None,
            "n": None,
            "s": None,
            "pile": None,
            "keys_shape": None,
            "keys_data_type": None,
            "values_shape": None,
            "values_data_type": None,
        }

        if url is not None:
            with open(url, 'r') as f:
                _conf = json.load(f)

                if not isinstance(_conf, dict):
                    die('Invalid configuration file!')

                for key in _conf:
                    if key == 'groups':
                        self.params['groups'].update(_conf['groups'])

                    else:
                        self.params[key] = _conf[key]

    def __getitem__(self, key):
        return self.params[key]

    def __str__(self):
        name_line = "<<< expam configuration file: %s >>>\n" % self.params['name']
        data_line = "phylogeny\t-->\t%s\nk\t\t-->\t%s\nn\t\t-->\t%s" % (self.params['phylogeny_path'],
                                                                    self.params['k'], self.params['n'])
        par_line = "sketch\t\t-->\t%s\npile\t\t-->\t%s" % (self.params['s'], self.params['pile'])

        group_fmt = "\n----------------\ngroup name: %s\n\tk\t\t-->\t%s\n\tsketch\t\t-->\t%s\n\tsequences\t-->\t%d\n"
        group_line = "\n".join([
            group_fmt % (name, group['k'], group['s'], len(group['sequence_files']))
            for name, group in self.params['groups'].items()
        ])

        return "\n".join([name_line, data_line, par_line, group_line])

    def get_paths(self):
        return [
            genome_path
            for group in self.params["groups"].values()
            for genome_path in group['sequence_files']
        ]

    def new_group(self, name, files=None, phylogeny_path=None, k=None, s=None):
        if name in self.params['groups']:
            raise ValueError("Group %s already exists!" % name)

        self.params['groups'][name] = {
            'sequence_files': [] if files is None else files,
            'phylogeny_path': phylogeny_path,
            'k': k,
            's': s,
        }

    def set(self, **kwargs):
        for arg in kwargs:
            if arg in self.params and arg != 'groups':
                if kwargs[arg] is not None:
                    self.params[arg] = kwargs[arg]

            else:
                raise ValueError("Unknown configuration category %s!" % arg)

    def group_set(self, group_name, **kwargs):
        for arg in kwargs:
            if arg in self.params['groups'][group_name] and arg != 'sequence_files' and kwargs[arg] is not None:
                self.params['groups'][group_name][arg] = kwargs[arg]

    def group_get(self, group_name):
        if group_name not in self.params['groups']:
            raise IndexError("Invalid group %s!" % group_name)

        group = self.params['groups'][group_name]
        attrs = {attr: group[attr] for attr in ('k', 's', 'sequence_files')}

        for req_param in ('k', 's'):
            if group[req_param] is None:
                if self.params[req_param]:
                    attrs[req_param] = int(self.params[req_param])

            else:
                attrs[req_param] = int(attrs[req_param])

        return attrs['k'], attrs['s'], attrs['sequence_files']

    def add_sequence(self, file_name, group=None):
        if group is None:
            group = 'default'

        if group not in self.params['groups']:
            self.new_group(group)

        if not isinstance(file_name, list):
            file_name = {file_name}
        else:
            file_name = set(file_name)

        # Check these sequences aren't already added.
        for current_group in self.params['groups']:
            if current_group != group:
                current_seqs = set(self.params['groups'][current_group]['sequence_files'])

                if file_name & current_seqs:
                    die("Some of these sequences are already in the database!")

        self.params['groups'][group]['sequence_files'].extend(file_name)

    def remove_sequence(self, file_name, group=None):
        if group is None:
            group = 'default'

        if isinstance(file_name, list):
            file_name = set(file_name)
        else:
            file_name = {file_name}

        sequences = set(self.params['groups'][group]['sequence_files'])
        sequences -= file_name

        self.params['groups'][group]['sequence_files'] = list(sequences)

    def remove_group(self, group=None):
        if group is None:
            self.params['groups']['default'] = {
                    'sequence_files': [],
                    'phylogeny_path': [],
                    'k': None,
                    's': None,
                }
        elif group in self.params['groups']:
            self.params['groups'].pop(group)
        else:
            raise Exception("No such group %s!" % group)

    def _dump(self):
        return json.dumps(self.params)

    def groups(self):
        return list(self.params['groups'].keys())

    def save(self, url=None):
        if url is None:
            if self.url is None:
                raise Exception("Save url required!")

            url = self.url

        with open(url, 'w') as f:
            json.dump(self.params, f)


class ControlCenter:
    def __init__(self, logging_dir, group_name=None, workers=None, child_statuses=None, phases=None,
                 phase_queues=None, child_queues=None, children=None, processors=None, transitions=None,
                 timeout=TIMEOUT):

        self.group_name = group_name
        self.workers = [] if workers is None else list(workers)

        self.error = False  # Error state.
        self._child_statuses = child_statuses  # Shared array.

        self.phases = [] if phases is None else list(phases)
        self.phase_queues = {} if phase_queues is None else dict(phase_queues)
        self.child_queues = child_queues

        # Connection objects.
        self._children = [] if children is None else list(children)

        # User defined methods.
        self._processors = {} if processors is None else dict(processors)
        self._transitions = {} if transitions is None else dict(transitions)
        self._shutdown = self.default_shutdown

        self._timeout = timeout

        # Logging to disk.
        self.logging_dir = logging_dir
        self.logger = new_logger(self.logging_dir, str(os.getpid()) + "_main")

    @property
    def phase(self):
        if not self.phases:
            return None

        else:
            return self.phases[0]

    def children_have_obeyed(self, *command_codes):
        if self._child_statuses is None:
            return True

        return all([
            child_status in command_codes
            for child_status in self._child_statuses
        ])

    @property
    def any_children_active(self):
        if self._child_statuses is None:
            return False

        return any([
            child_status == expam.processes.COMMAND_CODES["ACTIVE"]
            for child_status in self._child_statuses
        ])

    @property
    def children_are_dead(self):
        return self.children_have_obeyed(expam.processes.COMMAND_CODES["DEAD"])

    @classmethod
    def from_dict(cls, logging_dir, config):
        def check_int(value):
            if not isinstance(value, int):
                return False

            return True

        if not isinstance(config, dict):
            raise TypeError("Can only be instantiated using dictionary not %s!" % type(config))

        control_center = ControlCenter(logging_dir)

        required_args = [
            "phases",
            "layers",
            "name",
        ]
        optional_args = [
            "timeout",
        ]

        # Set optional arguments first.
        for arg in optional_args:
            if arg in config:

                if arg == "timeout":
                    if type(config[arg]) not in {float, int}:
                        raise ValueError("Parameter must be float or int not %s!" % type(config[arg]))

                    else:
                        control_center._timeout = config["timeout"]

        # Set required arguments.
        for arg in required_args:
            if arg not in config:
                raise ValueError("Required arg %s not in configuration!" % arg)

            else:
                if arg == "phases":
                    control_center.phases = list(config["phases"])

                elif arg == "layers":
                    layers_required_args = {
                        "class",
                        "parent",
                        "child",
                        "n"
                    }
                    current_layer = config["layers"]

                    if not isinstance(current_layer, dict) and "parent" in current_layer:
                        raise ValueError("Incorrect layer configuration!")

                    # Ensure that top layer is directly connected to the control center.
                    if current_layer["parent"] is not ControlCenter:
                        raise TypeError("Top level worker must be connected to ControlCenter not %s!"
                                        % current_layer["parent"])

                    parents = [control_center]

                    at_bottom = False
                    while not at_bottom:

                        if not isinstance(current_layer, dict):
                            raise ValueError("Layer configuration must be dictionary not %s!" % type(current_layer))

                        # Validate arguments.
                        for layer_arg in layers_required_args:
                            if layer_arg not in current_layer:
                                raise ValueError("Missing %s argument required for layer creation!" % layer_arg)

                            elif layer_arg == "n":
                                if not check_int(current_layer["n"]):
                                    raise TypeError("Layer parameter n must be int not %s!" % type(current_layer["n"]))

                            elif layer_arg == "class":
                                if not issubclass(current_layer["class"], expam.processes.JobWorker):
                                    raise TypeError("Process %s must be subclass of JobWorker!"
                                                    % str(current_layer["class"].__name__))

                            elif layer_arg == "child":
                                if current_layer["child"] is None:
                                    at_bottom = True

                        parents = control_center.add_layer(current_layer, parents)
                        current_layer = current_layer["child"]

                elif arg == "name":
                    control_center.group_name = str(config["name"])

        return control_center

    def add_layer(self, current_layer, parents):

        n_children = current_layer["n"]
        n_parents = len(parents)

        def make_shared_array():
            nonlocal n_children

            arr = Array(c_int8, n_children, lock=False)  # Only edited by children.
            for i in range(n_children):
                arr[i] = expam.processes.COMMAND_CODES["ACTIVE"]

            return arr

        def initialise_workers(statuses, timeout):
            nonlocal current_layer, n_children

            # Class of worker.
            worker = current_layer["class"]

            # Custom to initialise the worker.
            if "class_args" in current_layer:

                init_args = [
                    {
                        arg: value
                        if not (isinstance(value, tuple) and len(value) == n_children) else value[cid]
                        for arg, value in current_layer["class_args"].items()
                    }
                    for cid in range(n_children)
                ]

            else:
                init_args = [{} for _ in range(n_children)]

            # Insert default arguments.
            for cid in range(n_children):
                default_args = {
                    "status_index": cid,
                    "statuses": statuses,
                    "child_statuses": None,
                    "timeout": timeout,
                    "phases": self.phases,
                }

                init_args[cid] = {
                    **init_args[cid],
                    **default_args
                }

            return [
                worker(**init_args[cid])
                for cid in range(n_children)
            ]

        def make_pipes():
            nonlocal n_parents, n_children

            parent_cons = [[] for _ in range(n_parents)]
            child_cons = [[] for _ in range(n_children)]

            for i in range(n_parents):
                for j in range(n_children):
                    parent_con, child_con = Pipe()

                    parent_cons[i].append(parent_con)
                    child_cons[j].append(child_con)

            return parent_cons, child_cons

        def make_queues():
            nonlocal n_children

            return [
                Queue()
                for _ in range(n_children)
            ]

        current_statuses = make_shared_array()
        parent_cons, child_cons = make_pipes()
        work_queues = make_queues()

        for i, parent in enumerate(parents):
            parent._child_statuses = current_statuses
            parent._children = parent_cons[i]
            parent.child_queues = work_queues

        workers = initialise_workers(
            current_statuses,
            self._timeout
        )

        for i, worker in enumerate(workers):
            # Set communication routes.
            worker._fellow_statuses = current_statuses
            worker._parents = child_cons[i]
            worker.work_queue = work_queues[i]

        self.workers.append(workers)

        return workers

    def set_methods(self, methods_dict):

        if not isinstance(methods_dict, dict):
            raise TypeError("Config must be dictionary not %s!" % str(type(methods_dict)))

        for key, value in methods_dict.items():
            if value is None:
                continue

            if key == "shutdown":

                if not callable(value):
                    raise TypeError("Shutdown object must be callable!")

                self._shutdown = value

            elif key in self.phases:

                if not isinstance(value, dict):
                    raise TypeError("Method declaration must be dictionary not %s!"
                                    % str(type(value)))

                for method_name, method in value.items():
                    self.set_method(
                        method_name=method_name,
                        phase=key,
                        method=method
                    )

            else:
                raise ValueError("Unknown key in class config %s!" % key)

    def set_method(self, method_name, phase, method):
        methods = {
            "processor": self._processors,
            "transition": self._transitions,
        }

        if method_name not in methods:
            raise ValueError("Invalid method name %s!" % method_name)

        if method is not None:
            methods[method_name][phase] = method

    def default_shutdown(self):
        self.log("Control center shutting down.")

    def run(self):
        COMMAND_CODES = expam.processes.COMMAND_CODES

        self.start()

        try:

            # Phase loop.
            while self.phases:
                phase = self.phases[0]

                try:
                    self.log("Main filling queue at phase %s..." % self.phase)

                    # Processing loop.
                    while True:
                        # Check children for errors and any custom processing protocols.
                        self.process()

                        for child_queue in self.child_queues:
                            next_job = self.phase_queues[phase].get(block=True, timeout=self._timeout)

                            child_queue.put(next_job)

                except queue.Empty:
                    self.log("Exiting loop from MAIN...")

                except KeyError:

                    # Check there is work for this phase.
                    if phase not in self.phase_queues:
                        pass

                    else:
                        raise

                # Processing finished.
                self.log("Telling children to transition...")
                # Finalise data and transition to next phase.
                self.tell_children(
                    (
                        expam.processes.COMMAND_CODES["SLEEP"],
                        None
                    )
                )

                # Wait for response.
                while not self.children_have_obeyed(COMMAND_CODES["SLEEP"], COMMAND_CODES["DEAD"]):
                    self.process()

                # Transition now that children are transitioning.
                self.transition()  # Progresses current phase.

            # Shutdown phase.
            # Wait for processes to finish.

            while not self.children_have_obeyed(expam.processes.COMMAND_CODES["DEAD"]):
                self.check_children()

                self.timeout()

        except Exception as e:

            # Kill children before raising error.
            # Propagate error downward.
            self.finalise_error(
                -1,
                expam.processes.COMMAND_CODES["ERROR"],
                e
            )

            raise e

        finally:

            self.shutdown()

    def start(self):
        for layer in self.workers:
            for worker in layer:
                worker.start()

    def shutdown(self):
        # Default shutdown procedure.
        self._shutdown()

        # Gather processes.
        self.join()

    def process(self):
        self.check_children()

        if self.phase in self._processors:
            self._processors[self.phase]()

    def transition(self):
        self.log("Main transitioning...")

        if self.phase in self._transitions:
            self._transitions[self.phase]()

        self.phases.pop(0)
        self.log("Main now at phase %s." % ("end" if not self.phases else self.phase))

        # Wake up children and begin next phase.
        self.wake_children()
        self.log("Children are now awake, beginning next phase...")

    def wake_children(self):
        for cid in range(len(self._child_statuses)):
            self._child_statuses[cid] = expam.processes.COMMAND_CODES["ACTIVE"]

    def join(self):
        # Check that my queues are empty.
        all_queues = [*self.child_queues, *[phase_queue for phase_queue in self.phase_queues.values()]]

        for child_queue in all_queues:
            while True:
                try:
                    _ = child_queue.get(block=True, timeout=self._timeout)

                except queue.Empty:
                    break

        # Clean up workers.
        for layer in self.workers:
            for worker in layer:
                worker.join()

    def kill(self):
        for cid in range(len(self._children)):
            self.tell_child(
                cid=cid,
                message=(
                    expam.processes.COMMAND_CODES["DEAD"],
                    None
                )
            )

    def tell_children(self, message):
        for cid in range(len(self._children)):
            self.tell_child(cid, message)

    def tell_child(self, cid, message):
        self.child_queues[cid].put(message)

    def check_children(self):
        for cid in range(len(self._children)):
            self.check_child(cid)

    def check_child(self, cid):
        if self._children[cid].poll(self._timeout):
            err_code, err = self._children[cid].recv()
            self._children[cid].send(True)  # Read receipt.

            self.log("Error received at main!")

            # Propagate error downward.
            self.finalise_error(cid, err_code, err)

            # Processes should have died if an error has been raised.
            self.join()

            print("\n" + "-" * 30 + "\n  Exception raised to MAIN:\n" + "-" * 30)
            raise err

    def finalise_error(self, eid, code, err):
        self.log("MAIN propagating error...")
        self.error = True

        for cid in range(len(self._children)):
            if cid != eid:
                self._children[cid].send((code, err))

        # Wait for children to die.
        while not self.children_are_dead:
            self.timeout()

    def load_job(self, phase, job):
        if phase not in self.phases:
            raise ValueError("Undefined job phase %s!" % phase)

        # Initialise new phase.
        if phase not in self.phase_queues:
            self.phase_queues[phase] = Queue()

        self.phase_queues[phase].put(
            (
                expam.processes.COMMAND_CODES["ACTIVE"],
                job
            )
        )

    def timeout(self):
        time.sleep(self._timeout)

    def log(self, msg):
        self.logger.info(msg)


class ExpamProcesses(ControlCenter):

    def __init__(self, logging_dir, cons, out_dir, *args, **kwargs):
        super(ExpamProcesses, self).__init__(logging_dir, *args, **kwargs)

        # Define non-hierarchical connection to children.
        self.cons = cons
        self.out_dir = out_dir

        # Current UnionWorker array sizes.
        self.arr_sizes = {}

    @classmethod
    def from_method_dict(cls, connections, out_dir, logging_dir, config):

        # Configure base class from configuration dictionary.
        base_center = super(ExpamProcesses, cls).from_dict(logging_dir, config)

        base_arguments = {
            "group_name": "group_name",
            "workers": "workers",
            "child_statuses": "_child_statuses",
            "phases": "phases",
            "phase_queues": "phase_queues",
            "child_queues": "child_queues",
            "children": "_children",
            "processors": "_processors",
            "transitions": "_transitions",
            "timeout": "_timeout",
        }
        base_attributes = {
            attr: getattr(base_center, attr_reference)
            for attr, attr_reference in base_arguments.items()
        }

        # Create child class from this instance.
        control_center = ExpamProcesses(
            logging_dir=logging_dir,
            cons=connections,
            out_dir=out_dir,
            **base_attributes
        )

        # Initialise special methods.
        control_center.set_methods(
            {  # Custom methods for non-hierarchical communication between layers.
                "import": {
                    "processor": control_center.union_checker,
                    "transition": control_center.union_checker,
                },
                "map": {
                    "processor": None,
                    "transition": None,
                },
                "shutdown": control_center.union_saver,
            }
        )

        return control_center

    def union_checker(self):

        for i, con in enumerate(self.cons):
            if con.poll(self._timeout):

                cmd, data = con.recv()

                if cmd == "shm_extend":
                    shm_name, size, create = data

                    if shm_name is None:
                        # Create an allocation.
                        shm_all = shared_memory.SharedMemory(size=size, create=create)

                        shm_name = shm_all.name
                        shm_all.close()

                    else:
                        # Resize a current allocation.
                        shm_all = expam.stores.ResizableSharedMemory(shm_name)

                        shm_all.resize(size)
                        shm_all.close()

                    con.send(shm_name)

                elif cmd == "keys_size":
                    keys_shape = data

                    # Update known size.
                    self.arr_sizes[i] = keys_shape

                    # Set map size.
                    map_shape = keys_shape[:1]

                    # Create a store for its values.
                    shm_values_size = np.product(map_shape) * MAP_DATA_TYPE_SIZE
                    shm_all = shared_memory.SharedMemory(size=shm_values_size, create=True)

                    shm_name = shm_all.name
                    shm_all.close()

                    con.send(shm_name)

                else:
                    raise ValueError("Unrecognised command %s!" % cmd)

    def union_saver(self):

        if not self.error:

            # Calculate combined size.
            final_shape = tuple(
                fn([shape[i] for shape in self.arr_sizes.values()])
                for i, fn in enumerate((sum, max))
            )
            n_cols = final_shape[1]

            # Initialise database on disk.
            db_url = os.path.join(self.out_dir, DB_PATH)
            disk_array = expam.stores.TablesDb(
                db_url=db_url,
                shape=final_shape,
                keys_type=KEYS_DATA_TYPE,
                values_type=MAP_DATA_TYPE,
                create=True
            )
            disk_array.close()

            def add_tuples(*tuples):
                if tuples:
                    return tuple(
                        sum([tup[i] for tup in tuples])
                        for i in range(len(tuples[0]))
                    )

            # Tell each process to write to disk.
            current_shape = (0,) * n_cols

            for con in self.cons:

                while not con.poll(self._timeout):
                    # Don't cause hang due to un-propagated errors!
                    self.check_children()

                cmd, shape = con.recv()

                if cmd != "save":
                    raise ValueError("Unrecognised command %s!" % cmd)

                # Let process know where it comes in the file.
                con.send((current_shape[0], db_url))

                if shape is not None:
                    current_shape = add_tuples(current_shape, shape)

            # Save database parameters to config file.
            save_db_params(
                out_dir=self.out_dir,
                db_params={
                    "keys_shape": current_shape,
                    "keys_data_type": KEYS_DTYPE_STR,
                    "values_shape": current_shape[:1],
                    "values_data_type": MAP_DTYPE_STR,
                }
            )


def main(out_dir, genome_paths, phylogeny_path, k, n=None, n_extract=None,
         n_union=None, pile_size=None, file_name_mode=1, plot=False):
    logs_dir = os.path.join(out_dir, 'logs')

    # Initial time point.
    t0 = datetime.datetime.now()

    # Configure number of processes.
    if n_union is not None and n_extract is not None:
        n_union, n_extract = int(n_union), int(n_extract)

    elif n is not None:
        n_union, n_extract = distribute_processes(n)

    else:
        raise ValueError("Number of processes is can't be defined!")

    # Processing parameter details.
    num_cols = calculate_n_chunks(k)
    ranges = make_ranges(n_union, k)
    genome_paths, maxsize = sort_by_size(genome_paths)

    # Create shm_allocations for in-place kmer processing.
    shm_allocations = prepare_kmer_allocations(maxsize, num_cols, n_extract)

    # Import phylogeny.
    genome_ids, index = import_phylogeny(phylogeny_path)
    # Create LCA matrix.
    format_name = expam.sequences.format_name
    node_to_index = {node.name: i for i, node in enumerate(index.pool) if i > 0}
    lca_matrix = make_lca_matrix(index, node_to_index)

    # Prepare pipes between main and union workers.
    main_cons, union_cons = make_pipes(n_union)
    # Lock system to prevent concurrent access to HDF5 file.
    lock_value = Value(c_uint64, lock=True)

    # Multiprocessing configuration.
    mp_config = {
        "name": "expam",  # Job system title.
        "phases": [  # Two-pass algorithm.
            "import",  # Import sequences and take union.
            "map"  # Map kmer to LCA.
        ],
        "layers": {  # Top layer of child processes - extract sequence & kmers.
            "class": expam.processes.ExtractWorker,
            "class_args": {  # Positional arguments to init class.
                "k": k,
                "shm_kmer_params": shm_allocations,
                "kmer_ranges": ranges,
                "node_map": node_to_index,
                "out_dir": out_dir,
                "logging_dir": logs_dir,
            },
            "parent": ControlCenter,  # Receive work from main process.
            "child": {
                "class": expam.processes.UnionWorker,  # Take union of sets of kmers.
                "class_args": {  # Each either length n_union or 1.
                    "main_con": union_cons,
                    "lca_matrix": lca_matrix,
                    "pile_size": pile_size,
                    "logging_dir": logs_dir,
                    "lock_value": lock_value,
                },
                "parent": expam.processes.ExtractWorker,  # Receive work from ExtractWorkers.
                "child": None,  # Bottom of process hierarchy.
                "n": n_union,  # Number of processes.
            },
            "n": n_extract,
        },
        "timeout": TIMEOUT,  # Time to block for when checking queues.
    }

    process_network = ExpamProcesses.from_method_dict(main_cons, out_dir, logs_dir, mp_config)

    # Load files for importing stage.
    for file_dir in genome_paths:
        process_network.load_job(
            phase="import",
            job=file_dir
        )

    process_network.run()

    # Plot timing data.
    if plot:
        plot_data(logs_dir, t0)

    # Save coordinate names.
    node_names = [node.name for node in index.pool]
    with open(os.path.join(out_dir, COORDS_PATH), "w") as f:
        json.dump(node_names, f)

    # Save LCA matrix.
    np.save(os.path.join(out_dir, LCA_MATRIX_PATH), lca_matrix)

    # Concatenate accession ID files.
    concatenate_accession_ids(os.path.join(out_dir, "phylogeny"))


def save_db_params(out_dir, db_params):
    conf_dir = os.path.join(out_dir, 'expam.conf')

    conf = JSONConfig(url=conf_dir)
    conf.set(**db_params)
    conf.save()


def distribute_processes(n):
    n_union = math.floor(UNION_RATIO * n)
    n_extract = n - n_union

    return n_union, n_extract


def make_ranges(n, k):
    # Calculate hash offset for the k value.
    maxvalue = maximum_hash(k)

    # Interpolate between offset and maxvalue.
    dx = maxvalue // n
    return {
        i: (i * dx, (i + 1) * dx) if i < n - 1
        else (i * dx, maxvalue)
        for i in range(n)
    }


def maximum_hash(k):
    if k < CHUNK_SIZE:
        return 4 ** k - 1
    else:
        return 4 ** CHUNK_SIZE - 1


def sort_by_size(dirs):
    _file_suffixes = expam.sequences.COMP_PARSE
    _suffix_check = expam.sequences.check_suffix

    def _gzip_size(file_dir):
        file_stats = subprocess.run(["gzip", "-l", file_dir], capture_output=True)
        file_meta = [entry for entry in file_stats.stdout.split(b' ') if entry]
        return file_meta[5]

    def _fna_size(file_dir):
        return os.stat(file_dir).st_size

    def _get_file_size(file_dir):
        for suffix in _file_suffixes:
            if _suffix_check(file_dir, suffix):
                return _gzip_size(file_dir)

        return _fna_size(file_dir)

    max_filename = max([len(f) for f in dirs])
    file_dt = np.dtype([
        ("filename", "S%d" % max_filename),
        ("filesize", np.int64)
    ])

    # Turn into structured array.
    file_array = np.array([
        (file_dir, _get_file_size(file_dir))
        for file_dir in dirs
    ], dtype=file_dt)

    # Sort by file size.
    file_array.sort(order="filesize")

    max_size = file_array["filesize"][-1]
    ordered_files = [
        filename.decode("utf8")
        for filename in file_array["filename"].tolist()
    ]

    # Assuming each char in the file is a base.
    return ordered_files, max_size


def prepare_kmer_allocations(rows, cols, n_processes):
    allocation_params = ()
    allocation_size = rows * cols * KEYS_DATA_TYPE_SIZE
    allocation_shape = (rows, cols)

    for _ in range(n_processes):
        # Create and 0 allocation.
        next_shm_allocation = shared_memory.SharedMemory(
            create=True,
            size=allocation_size
        )

        next_arr = np.ndarray(
            shape=allocation_shape,
            dtype=KEYS_DATA_TYPE,
            buffer=next_shm_allocation.buf
        )
        next_arr[:] = 0

        # Pass on allocation details to children.
        allocation_params += ((
                                  next_shm_allocation.name,
                                  allocation_shape
                              ),)

        next_shm_allocation.close()

    return allocation_params


def calculate_n_chunks(k):
    return k // CHUNK_SIZE + (k % CHUNK_SIZE != 0)


def import_phylogeny(phylogeny_path):
    print("Importing phylogeny...")

    # Load the phylogeny into an Index.
    genome_names, index = expam.tree.Index.load_newick(phylogeny_path)
    return genome_names, index


def yield_coordinates(index):
    n = len(index.pool)

    for i in range(1, n):
        for j in range(1, i):
            yield i, j, index.pool[i].coordinate, index.pool[j].coordinate


def compute_lca(i, j, coord_one, coord_two):
    return i, j, expam.tree.propose_lca(coord_one, coord_two)


def make_lca_matrix(index, node_to_index):
    print("Creating LCA matrix...")

    def get_children(fixed_node, flexible_index_list):
        nonlocal index, matrix, node_to_index

        i = 0
        while i < len(flexible_index_list):
            for child in index.pool[flexible_index_list[i]].children:
                m = min(fixed_node, node_to_index[child])
                n = max(fixed_node, node_to_index[child])

                if matrix[n, m] == 0 and n != m:
                    flexible_index_list.append(node_to_index[child])

            i += 1

    n = len(index.pool)
    matrix = np.zeros((n, n), dtype=np.uint16)

    for i in range(1, n):
        subject = index.pool[i].coordinate
        subject_name = index.pool[i].name

        for j in range(1, i):
            if matrix[i, j] > 0:
                continue

            target = index.pool[j].coordinate
            target_name = index.pool[j].name

            lca_coordinate = expam.tree.propose_lca(subject, target)
            lca = index.coord(coordinate=lca_coordinate).name

            left_group = [i]
            right_group = [j]

            if lca not in [subject_name, target_name]:
                # Subject and target are on separate lineages.
                get_children(j, left_group)
                get_children(i, right_group)

            elif lca == subject_name:
                # Target is contained in the clade rooted at subject.
                get_children(i, right_group)

            elif lca == target_name:
                # Subject is contained in the clade rooted at target.
                get_children(j, left_group)

            for p in left_group:
                for q in right_group:
                    m = min(p, q)
                    n = max(p, q)

                    if n != m and matrix[n, m] == 0:
                        matrix[n, m] = node_to_index[lca]

    return matrix


def make_pipes(n):
    parent_cons, child_cons = [], []

    i = 0
    while i < n:
        parent_con, child_con = Pipe()

        parent_cons.append(parent_con)
        child_cons.append(child_con)

        i += 1

    return tuple(parent_cons), tuple(child_cons)


def plot_data(logs_dir: str, t0: datetime.datetime) -> None:
    def _get_time(string):
        nonlocal t0

        date, current_time, am_pm = string.split(" ")[:3]
        mon, day, yr = (int(v) for v in date.split("/"))
        hr, mn, sec = (int(v) for v in current_time.split(":"))

        if "PM" in am_pm and hr < 12:
            hr += 12
        elif "AM" in am_pm and hr == 12:
            hr = 0

        time_point = datetime.datetime(yr, mon, day, hr, mn, sec)

        return (time_point - t0).total_seconds()

    def _get_data(string):
        return float(
            (string.split(" took ")[1])[:-3]
        )

    job_type_map = {  # For labels.
        "request_extension": "remap",
        "disjoint": "disjoint",
        "collapse_pile": "fill",
        "import_sequence": "kmers",
        "send_kmers": "send",
    }
    job_colours = {  # For plotting.
        "request_extension": "red",
        "disjoint": "blue",
        "collapse_pile": "green",
        "import_sequence": "orange",
        "send_kmers": "black",
    }

    process_type_jobs = {
        "_extract": [
            "import_sequence",
            "send_kmers",
        ],
        "_union": [
            "request_extension",
            "disjoint",
            "collapse_pile"
        ],
    }

    fig, ax = plt.subplots(figsize=(20, 15))
    ax.grid()

    log_urls = [
        os.path.join(logs_dir, log_name)
        for log_name in expam.sequences.ls(logs_dir, ext=".log")
        if "_main.log" not in log_name
    ]

    # Plot data one child process at a time.
    for log_url in log_urls:
        process_type = "_extract" if "_extract.log" in log_url else "_union"

        # Data for plotting.
        data = {job_name: ([], []) for job_name in process_type_jobs[process_type]}

        with open(log_url, "r") as f:
            log_data = f.readlines()

        # Read data line by line and append data.
        for i, line in enumerate(log_data):
            for job_name in process_type_jobs[process_type]:

                if job_name in line:
                    data[job_name][0].append(_get_time(line) / 3600)
                    data[job_name][1].append(_get_data(line) / 60)

        # Plot data on same graph.
        for job_name in process_type_jobs[process_type]:
            x, y = np.array(data[job_name][0]), np.array(data[job_name][1])
            ax.plot(
                x,
                y,
                label=job_type_map[job_name],
                marker="x",
                color=job_colours[job_name],
            )

            # Don't allow duplicate labels.
            job_type_map[job_name] = ""

    ax.set(
        xlabel="Time since start [hrs]",
        ylabel="Time taken for task [mins]",
        title="Time taken to do tasks"
    )

    ax.legend(loc='best')
    fig.savefig(os.path.join(logs_dir, "timing.png"))


def timeit(func):
    def inner(*args, **kwargs):
        with Timer() as timer:
            returns = func(*args, **kwargs)

        print(f'@timeit {func.__name__} of {func.__module__}: ', timer)
        return returns

    return inner


def log_timer(func):
    # Time function and log.
    def inner(*args, **kwargs):

        loggers = logging.root.manager.loggerDict
        my_logger = None  # Null case.

        # Appropriate logger will be identified by process id.
        pid = os.getpid()

        for name in loggers:
            if str(pid) in name:
                my_logger = loggers[name]

                break

        if my_logger is None:
            raise Exception("Logger not found!")

        with Timer() as timer:
            returns = func(*args, **kwargs)

        my_logger.info(f'@timeit {func.__name__} of {func.__module__} took {str(timer)}.')
        return returns

    return inner


def new_logger(logs_dir, logger_name):
    """
    Create a fresh Logger instance that doesn't propagate
    to the root logger. Used when a new process is created.

    :param logs_dir:
    :param logger_name:
    :return:
    """
    log_file_url = os.path.join(logs_dir, logger_name + ".log")

    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)
    logger.propagate = False  # Don't propagate to root logger.

    if not logger.hasHandlers():  # Don't duplicate logger file handlers.
        # Let each process log to a unique file.
        fh = logging.FileHandler(log_file_url)
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(LOG_FORMAT)

        logger.addHandler(fh)

    return logger


def concatenate_accession_ids(f_url):
    def get_data(file_url):
        with open(file_url, "r") as f:
            data = f.read()

        # Delete temporary data file.
        os.remove(file_url)

        return data

    def get_files(folder_url):
        return [
            file_url
            for file_url in expam.sequences.ls(folder_url)
            if "accession_ids_" in file_url
        ]

    id_data = [
        get_data(os.path.join(f_url, file_name))
        for file_name in get_files(f_url)
    ]

    with open(os.path.join(f_url, "accession_ids.csv"), "w") as f:
        f.write("\n".join(id_data))



import io
import os
import re
from math import sqrt
from multiprocessing.shared_memory import SharedMemory
from os import ftruncate

import numpy as np
import tables as tb
from tables import UInt64Atom, UInt16Atom
from expam.sequences import format_name, get_opener

CHUNK_SIZE = int(1e9)  # 1Gb sequence file chunking.


class MaskStore:
    def __init__(self):
        self.store = {}

    def __setitem__(self, key, value):
        # Place value in store.
        self.store[key] = value

    def __getitem__(self, key):
        return self.store[key]

    def __delitem__(self, key):
        del self.store[key]


class SequenceStore:
    def __init__(self, k, out_dir):
        self._store = {}
        self.stream = io.BytesIO()
        self.k = k
        self.out_dir = out_dir

        self.accession_ids = {}

    def add_sequence(self, file_path):
        # Check it is a valid file.
        if not os.path.isfile(file_path):
            raise ValueError("Directory %s is not a file!" % file_path)

        print("Extracting sequences from %s..." % file_path)

        # Determine opening mode based on compression of sequence file.
        file_name = format_name(file_path)
        file_name, (mode, opener) = get_opener(file_name)

        with opener(file_path, mode=mode) as f:
            # Extract the accession id.
            metadata = f.readline()
            assert b'>' in metadata

            accession_id, tax_id = self.extract_accession_id(metadata)

            if accession_id is None and tax_id is None:
                raise ValueError("Couldn't identify sequence %s!" % file_name)

            for line in f:
                line = line.strip()
                # Skip information line at start of FASTA format.
                if b'>' in line:
                    self.stream.write(b'N')
                else:
                    self.stream.write(line)

        # Convert to numeric sequence and put in store.
        sequence_view = self.stream.getvalue()
        self._store[file_name] = sequence_view

        # Save accession ID for this file.
        self.accession_ids[file_name] = [accession_id, tax_id]

        # Clear stream.
        self.stream.seek(0)
        self.stream.truncate()

        return file_name, len(self._store[file_name])

    def __getitem__(self, key):
        # If it is a branch, this is a valid request,
        # but no information from the Sequence Store is required.
        if key.isdigit():
            return None
        else:
            return self._store.__getitem__(key)

    def __setitem__(self, key, value):
        self._store.__setitem__(key, value)

    def __delitem__(self, key):
        self._store.__delitem__(key)

    def keys(self):
        return self._store.keys()

    @staticmethod
    def extract_accession_id(string):
        def try_convert(byte_string):
            try:
                return byte_string.decode("utf8")

            except AttributeError:
                return byte_string

        accn_id = re.findall(b'(?:^\s*\>)(.*?)(?:\s+|\|)', string)
        tax_id = re.findall(b'(?:\|kraken:taxid\||\|taxid\|)(\d+)(?:\s+|\|)', string)

        accn_id = None if not accn_id else try_convert(accn_id[0])
        tax_id = None if not tax_id else try_convert(tax_id[0])

        return accn_id, tax_id

    def save_accession_ids(self, suffix=""):
        id_data = "\n".join([
            "%s,%s,%s" % (name, keys[0], keys[1])
            for name, keys in self.accession_ids.items()
        ])

        with open(os.path.join(self.out_dir, "phylogeny", "accession_ids_%s.csv" % suffix), "w") as f:
            f.write(id_data)


class ResizableSharedMemory(SharedMemory):
    def __init__(self, name=None, create=False, size=0):
        super(ResizableSharedMemory, self).__init__(name, create, size)

    @property
    def size(self):
        """Size in bytes."""
        return self._size

    @size.setter
    def size(self, val):
        self._size = val

    def resize(self, newsize):
        fd = self._fd

        try:
            ftruncate(fd, newsize)
            return 1, newsize
        except OSError as e:
            return 0, e

    def remap(self, newsize):
        if newsize != self.size:
            del self._buf  # Clear buffer exposure.
            self._mmap.resize(newsize)  # Remap memory.
            self._buf = memoryview(self._mmap)  # Re-expose buffer.
            # Update size.
            self.size = newsize


class TablesDb:
    def __init__(self, db_url, shape=None, keys_type=None, values_type=None, create=False, edit=False):
        self.url = db_url

        self.keys_shape = shape
        self.values_shape = None if shape is None else tuple((shape[0],))

        self.keys_type = keys_type
        self.values_type = values_type

        self.edit = True if (create or edit) else False

        self.keys_offset = 0
        self.values_offset = 0

        filters = tb.Filters(complevel=5, complib="zlib")

        if create:
            if any([shape is None, keys_type is None, values_type is None]):
                raise ValueError("Parameters not provided for creation!")

            self.fh = tb.open_file(db_url, mode="w")

            self.keys = self.fh.create_carray(self.fh.root, "keys", UInt64Atom(),
                                              self.keys_shape, filters=filters)
            self.values = self.fh.create_carray(self.fh.root, "values", UInt16Atom(),
                                                self.values_shape, filters=filters)

        else:
            if edit:
                self.fh = tb.open_file(db_url, mode="r+")

                self.keys = self.fh.root.keys
                self.values = self.fh.root.values

            else:
                self.fh = tb.open_file(db_url, mode="r")

    def _write(self, src, dest, offset):
        if not self.edit:
            raise IOError("Opened in read mode!")

        src_length = src.shape[0]
        dest[offset:offset + src_length] = src[:]

        return src_length

    def write_keys(self, arr, offset=0):
        return self._write(src=arr, dest=self.keys, offset=offset)

    def write_values(self, arr, offset=0):
        return self._write(src=arr, dest=self.values, offset=offset)

    def read_keys(self, arr):
        if self.edit:
            raise IOError("Opened in create mode!")

        self.fh.root.keys.read(out=arr)

    def read_values(self, arr):
        if self.edit:
            raise IOError("Opened in create mode!")

        self.fh.root.values.read(out=arr)

    def close(self):
        self.fh.close()


class LTMatrix:
    def __init__(self, n):
        self._width = n-2
        self._len = int((n-2) * (n-1) / 2)
        self._arr = np.zeros(self._len, dtype=np.uint16)

    def __getitem__(self, index):
        if len(index) != 2:
            raise IndexError("Can only take indices of length 2!")

        i, j = index
        if i == 0 or j == 0:
            return 0

        if i == j:
            return i

        return self._arr[self._calc_index(i - 1, j - 1)]

    def __setitem__(self, index, value):
        if len(index) != 2:
            raise IndexError("Can only take indices of length 2!")

        i, j = index
        if (i == j) or any((i == 0, j == 0)):
            raise IndexError("Can only edit lower triangular numbers!")

        self._arr[self._calc_index(i - 1, j - 1)] = value

    def __str__(self):
        string = ""

        i = 0
        for a in range(1, self._width + 1):
            string += "\t".join(str(j) for j in self._arr[i:i+a]) + "\n"
            i += a

        return string.strip()

    @staticmethod
    def _calc_index(i, j):
        if i < j:
            raise IndexError("Not lower triangular index!")

        return int((0.5 * (i-1) * i) + j)

    def save(self, out_dir):
        np.save(out_dir, self._arr)

    @classmethod
    def load(cls, in_dir):
        arr = np.load(in_dir)
        l = len(arr)
        n = int(0.5 * (3 + sqrt(1 + (8*l))))

        obj = cls(n)
        obj._arr[:] = arr[:]

        return obj



#cython: infer_types=True
#cython: boundscheck=False
#cython: wraparound=False
#cython: nonecheck=False
#cython: cdivision=True

cimport cython
from libc.stdlib cimport malloc, free
cimport numpy as np
import numpy as np

ctypedef np.uint8_t uint8_t
ctypedef np.uint32_t uint32_t
ctypedef np.uint64_t uint64_t
ctypedef unsigned long long uint64
ctypedef long long int int64

cdef extern from "kmers.c":
    cdef uint64_t raw_kmers(const unsigned char *sequence, int k, uint64_t *arr)
    cdef void fill_kmers(const unsigned char *sequence, int k, uint64_t n, uint64_t *arr)

cdef extern from "jellyfish.c":
    cdef uint64_t word_reverse_complement(uint64_t word, int k)

cpdef void reverse_complement(uint64_t[::1] kmers, uint64_t[::1] r_kmers, int k):
    cdef:
        int j
        int m = kmers.shape[0]-1, b = 2 * ((k % 32) + (32 * (k % 32 == 0)))
        uint64_t r, mask = ((2**b) - 1)

    for j in range(m):
        if j == m-1:
            r_kmers[j] = word_reverse_complement(kmers[j+1], k)
        else:
            r_kmers[j] = word_reverse_complement(kmers[j+1], 32)

    if m > 1:
        # Swap order of chunks.
        for j in range(m // 2):
            r = r_kmers[m-(j+1)]
            r_kmers[m-(j+1)] = r_kmers[j]
            r_kmers[j] = r

        # Fix gaps in chunks.
        for j in range(m-1):
            r_kmers[j] <<= (64 - b)
            r = r_kmers[j+1] >> b
            r_kmers[j] |= r

        r_kmers[m-1] &= mask

cdef bint kmer_less(uint64_t[::1] a, uint64_t[::1] b):
    cdef int i, l = a.shape[0]

    for i in range(l):
        if a[i] < b[i+1]:
            return 1
        elif a[i] > b[i+1]:
            return 0

    return 0

cdef void canonical(uint64_t[:, ::1] kmers, int k):
    cdef:
        int i, l = kmers.shape[0], m = kmers.shape[1]-1
        uint64_t[::1] kmer, r_comp = np.ndarray(m, dtype=np.uint64)

    for i in range(l):
        kmer = kmers[i]
        reverse_complement(kmer, r_comp, k)

        if kmer_less(r_comp, kmer):
            kmers[i, 1:] = r_comp[:]

def get_raw_kmers(const unsigned char[::1] sequence, int k, uint64_t[:, ::1] arr):
    """Extract k-mers from sequence into the last (k//32+!!(k%32!=0)) columns of arr.
    The indices of these kmers (with respect to sequence) are stored in the
    first column of arr.

    :param sequence: Nucleotide sequence.
    :type sequence: bytes
    :param k: k value for analysis.
    :param arr: Array where kmers and their indices will be placed.
    :return: The number of rows of arr where kmers have been placed.
    :rtype: int

    """
    cdef uint64_t n_kmers

    n_kmers = raw_kmers(&sequence[0], k, &arr[0, 0])
    canonical(arr, k)

    return n_kmers

def bytes_intersect(const unsigned char[:] p, const unsigned char[:] q):
    cdef int l = min(len(p), len(q))

    for i in range(l):
        if p[i] != q[i]:
            return bytes(p[:i])
    else:
        return bytes(p[:l])

def reverse_complement_combine(const unsigned char[:] old, const unsigned char[:] new):
    cdef:
        int l, n, d

    l = len(old)
    n = len(new)
    d = n + l

    view = np.frombuffer(old, dtype='S1')
    dest = np.ndarray(d + 1, dtype='S1')

    for i in range(l):
        dest[i] = view[i]

    dest[l] = b'N'

    for i in range(n):
        if new[i] == b'A':
            dest[d-i] = b'T'
        elif new[i] == b'C':
            dest[d-i] = b'G'
        elif new[i] == b'G':
            dest[d-i] = b'C'
        elif new[i] == b'T':
            dest[d-i] = b'A'
        else:
            dest[d-i] = new[i]

    return dest.tobytes()

def init_index(const unsigned char[::1] sequence, int k, uint64_t[:, ::1] arr):
    """Equivalent to get_raw_kmers, however valid kmer indices are first found
    and then the kmers corresponding to these indices are then extracted
    independently. (Prefer to use get_raw_kmers.)

    :param sequence: Nucleotide sequence.
    :type sequence: bytes
    :param k: k value for analysis.
    :param arr: Array where kmers and their indices will be placed.
    :return: The number of rows of arr where kmers have been placed.
    :rtype: int

    """
    cdef uint64_t i

    i = start_mask(sequence, k, arr[:, 0])
    fill_kmers(&sequence[0], k, i, &arr[0, 0])

    canonical(arr, k)

    return i

def remove_duplicates(uint64_t[:, :] arr):
    """Remove (in-place) duplicate elements of a sorted arr. Equality
    is determined by values in the second column onwards.

    :param arr: Array of values.
    :return: Number of unique elements remaining.
    :rtype: int

    """
    cdef:
        uint64 arr_rows, arr_cols
        uint64 i, j, x

    arr_rows = arr.shape[0]
    arr_cols = arr.shape[1]

    x = 1

    for i in range(1, arr_rows):

        # Check last indices first as they're more likely to not agree.
        for j in range(arr_cols - 1, 0, -1):    # Don't look at index column.
            if arr[i, j] != arr[i - 1, j]:

                arr[x, :] = arr[i, :]
                x += 1

                break

    return x

def to_32bit(uint64_t[:] arr):
    return np.ascontiguousarray(arr, dtype=np.uint32)

def import_mask(uint32_t[::1] data, uint64_t[:] dest):
    move_32bit(data, dest)

def map_kmers(const unsigned char[::1] sequence, int k, uint64_t[:, ::1] arr):
    """Given an array whose first column is filled with indices of elements in
    some sequence, extract the kmers starting at each of these indices
    into the remaining columns of arr.

    :param sequence: Nucleotide sequence.
    :type sequence: bytes
    :param k: k value.
    :param arr: Array containing indices and space for kmers.
    :return: None, operations are done in-place.

    """
    cdef uint64_t n

    n = arr.shape[0]
    fill_kmers(&sequence[0], k, n, &arr[0, 0])
    canonical(arr, k)

def extract_kmers_from_read(const unsigned char[::1] read, int k):
    """Extract the kmers from a sequence into a new array.

    :param read: Short (~150bp) nucleotide sequence.
    :param int k: k value for analysis.
    :return: Array of kmers.
    :rtype: np.ndarray

    """
    cdef:
        uint64_t length, chunks, n_kmers
        uint64_t[:, ::1] kmers

    chunks = k // 32 + (k % 32 != 0)
    length = len(read)

    kmers = np.ndarray(
        shape=(length, chunks + 1),
        dtype=np.uint64
    )

    n_kmers = raw_kmers(&read[0], k, &kmers[0, 0])
    canonical(kmers, k)

    return kmers[:n_kmers]

cdef uint64_t start_mask(const unsigned char[::1] sequence, int k, uint64_t[:] index):
    cdef:
        uint64_t sequence_length
        uint64_t i, j, m
        char *valid_bits

    sequence_length = sequence.shape[0]
    i = j = 0

    valid_bits = <char *> malloc(sequence_length * sizeof(char))

    if not valid_bits:
        raise MemoryError

    try:
        # Each bit is a valid candidate index.
        for i in range(sequence_length):
            valid_bits[i] = 1

        for i in range(sequence_length):
            if sequence[i] not in b"ACGT":

                # Don't invalid bits using negative index.
                m = 0 if i + 1 - k < 0 else i + 1 - k

                for j in range(m, i+1):
                    valid_bits[j] = 0

        # Final edge.
        for j in range(k-1):
            valid_bits[-j] = 0

        # Convert bits into list of indices.
        j = 0
        for i in range(sequence_length):
            if valid_bits[i] == 1:
                index[j] = i
                j += 1

    finally:
        free(valid_bits)

    return j

cdef void move_32bit(uint32_t[::1] source, uint64_t[:] dest):
    cdef int i, length

    length = source.shape[0]

    for i in range(length):
        dest[i] = <uint64_t>source[i]
